/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: Builder.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.hierarchy;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;
import it.tidalwave.util.Id;
import it.tidalwave.util.IdFactory;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: Builder.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
public abstract class Builder<T>
  {
    @CheckForNull
    protected String displayName;

    protected boolean canBeRenamed = false;

    @CheckForNull
    protected Id id;

    @CheckForNull
    protected IdFactory idFactory;

    @CheckForNull
    protected Object boundObject;

    protected final List<Object> withObjects = new ArrayList<Object>();

    /*******************************************************************************************************************
     *
     * Specifies the display name of the object to build.
     *
     * @param  displayName   the display name
     * @return               itself
     *
     ******************************************************************************************************************/
    @Nonnull
    public Builder<T> withDisplayName (final @Nonnull String displayName)
      {
        this.displayName = displayName;
        return this;
      }

    /*******************************************************************************************************************
     *
     * Specifies that the object to build can be renamed.
     *
     * @return               itself
     *
     ******************************************************************************************************************/
    @Nonnull
    public Builder<T> thatCanBeRenamed()
      {
        return thatCanBeRenamed(true);
      }

    /*******************************************************************************************************************
     *
     * Specifies that the object to build can be renamed.
     *
     * @param  canBeRenamed  whether can be renamed
     * @return               itself
     *
     ******************************************************************************************************************/
    @Nonnull
    public Builder<T> thatCanBeRenamed (final boolean canBeRenamed)
      {
        this.canBeRenamed = canBeRenamed;
        return this;
      }

    /*******************************************************************************************************************
     *
     * Specifies the id of the object to build.
     *
     * @param   object           the id
     * @return               itself
     *
     ******************************************************************************************************************/
    @Nonnull
    public Builder<T> withId (final @Nonnull String id)
      {
        return withId(new Id(id));
      }

    /*******************************************************************************************************************
     *
     * Specifies the id of the object to build.
     *
     * @param   object           the id
     * @return               itself
     *
     ******************************************************************************************************************/
    @Nonnull
    public Builder<T> withId (final @Nonnull Id id)
      {
        this.id = id;
        return this;
      }

    /*******************************************************************************************************************
     *
     * Specifies the {@link IdFactory} of the object to build. It will be used to generate the ids of its children.
     *
     * @param   idFactory    the IdFactory
     * @return               itself
     *
     ******************************************************************************************************************/
    @Nonnull
    public Builder<T> withIdFactory (final @Nonnull IdFactory idFactory)
      {
        this.idFactory = idFactory;
        return this;
      }

    /*******************************************************************************************************************
     *
     * Specifies the {@link Object} that will be logically bound with the object to create.
     *
     * @param   boundObject  the bound object
     * @return               itself
     *
     ******************************************************************************************************************/
    @Nonnull
    public Builder<T> boundTo (final @Nonnull Object boundObject)
      {
        this.boundObject = boundObject;
        return this;
      }

    /*******************************************************************************************************************
     *
     * Specifies and extra capability.
     *
     * @param   object       the capability
     * @return               itself
     *
     ******************************************************************************************************************/
    @Nonnull
    public Builder<T> with (final @Nonnull Object object)
      {
        this.withObjects.add(object);
        return this;
      }

    /*******************************************************************************************************************
     *
     * Builds the object.
     *
     * @return   the built object
     *
     ******************************************************************************************************************/
    @Nonnull
    public abstract T build();
  }
