/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: HViewHItemCommonTestSupport.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.hierarchy;

import it.tidalwave.hierarchy.HView;
import java.util.ArrayList;
import java.util.List;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;
import it.tidalwave.util.AsException;
import it.tidalwave.util.Displayable;
import it.tidalwave.util.Id;
import it.tidalwave.util.MutableDisplayable;
import it.tidalwave.util.NotFoundException;
import static it.tidalwave.util.Displayable.Displayable;
import static it.tidalwave.hierarchy.HItem.Copiable;
import static it.tidalwave.hierarchy.HItem.Movable;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;
import static org.hamcrest.CoreMatchers.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: HViewHItemCommonTestSupport.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
public class HViewHItemCommonTestSupport<T extends HView>
  {
    protected T fixture;

    public static class BoundObject
      {
      }

    //////////////////////// Default Item
    @Test
    public void aDefaultItemMustHaveTheCorrectName()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   build();
        assertThat(item, is(notNullValue()));
        assertThat(item.as(Displayable).getDisplayName(), is("My Item 1"));
      }

    @Test
    public void aDefaultItemCannotBeRenamed()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   build();
        assertThat(item, is(notNullValue()));
        assertThat(item.as(Displayable), is(not(instanceOf(MutableDisplayable.class))));
      }

    @Test
    public void aDefaultItemMustHaveAnId()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   build();
        assertThat(item, is(notNullValue()));
        assertThat(item.getId(), is(notNullValue()));
      }

    @Test
    public void aDefaultItemMustBeInitiallyEmpty()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   build();
        assertThat(item.findChildren().count(), is(0));
      }

    @Test(expected=AsException.class)
    public void aDefaultItemMustNotBeCopiable()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   build();
        item.as(Copiable);
      }

    @Test(expected=AsException.class)
    public void aDefaultItemMustNotBeMovable()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   build();
        item.as(Movable);
      }

    //////////////////////// Renamable Item
    @Test
    public void aRenamableItemMustHaveTheCorrectName()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   thatCanBeRenamed().
                                   build();
        assertThat(item, is(notNullValue()));
        assertThat(item.as(Displayable).getDisplayName(), is("My Item 1"));
      }

    @Test
    public void aRenamableItemCanBeRenamed()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   thatCanBeRenamed().
                                   build();
        assertThat(item, is(notNullValue()));
        item.as(MutableDisplayable.class).setDisplayName("New Item Name");
        assertThat(item.as(Displayable).getDisplayName(), is("New Item Name"));
      }

    @Test
    public void aRenamableItemMustHaveAnId()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   thatCanBeRenamed().
                                   build();
        assertThat(item, is(notNullValue()));
        assertThat(item.getId(), is(notNullValue()));
      }

    @Test
    public void aRenamableItemMustBeInitiallyEmpty()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   thatCanBeRenamed().
                                   build();
        assertThat(item.findChildren().count(), is(0));
      }

    @Test(expected=AsException.class)
    public void aRenamableItemMustNotBeCopiable()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   thatCanBeRenamed().
                                   build();
        item.as(Copiable);
      }

    @Test(expected=AsException.class)
    public void aRenamableItemMustNotBeMovable()
      {
        final HItem item = fixture.createChild().
                                   withDisplayName("My Item 1").
                                   thatCanBeRenamed().
                                   build();
        item.as(Movable);
      }

    ////// MISC
    @Test
    public void mustCreateItemsWithDifferentIds()
      {
        final List<Id> ids = new ArrayList<Id>();

        for (int i = 0; i < 100; i++)
          {
            final HItem item = fixture.createChild().
                                       withDisplayName("My Item " + i).
                                       build();
            assertFalse(ids.contains(item.getId()));
            ids.add(item.getId());
          }
      }

    @Test
    public void mustBeAbleToFindItemById()
      throws NotFoundException
      {
        final HItem item1 = fixture.createChild().
                                    withDisplayName("My Item 1").
                                    build();
        final HItem item2 = fixture.createChild().
                                    withDisplayName("My Item 2").
                                    build();

        assertThat(fixture.findChildren().id(item1.getId()).result(), is(sameInstance(item1)));
        assertThat(fixture.findChildren().id(item2.getId()).result(), is(sameInstance(item2)));

        assertThat(fixture.findChildren().id("fakeId").count(), is(0));
      }

    @Test
    public void testFindOrCreateItemWithNoItem()
      {
        final HItem item = fixture.findOrCreateChild().withId("id1").
                                                       withDisplayName("the name").
                                                       build();
        assertThat(item.getId(), is(new Id("id1")));
      }

    @Test
    public void testFindOrCreateItemWithExistingItem()
      {
        final HItem item1 = fixture.createChild().withDisplayName("a name").build();
        final HItem item2 = fixture.findOrCreateChild().withId(item1.getId()).
                                                        withDisplayName("the name").
                                                        build();
        assertThat(item2, is(sameInstance(item1)));
        assertThat(item1.as(Displayable).getDisplayName(), is("a name"));
      }

    @Test
    public void mustUseTheDisplayableOfBoundObjectWhenNoExpicitDisplayableIsProvided()
      {
        final Displayable displayable = mock(Displayable.class);
        final Lookup lookup = Lookups.fixed(displayable);
        final Lookup.Provider boundObject = mock(Lookup.Provider.class);
        when(boundObject.getLookup()).thenReturn(lookup);

        final HItem item = fixture.createChild().boundTo(boundObject).build();
        assertThat(item.as(Displayable), is(sameInstance(displayable)));
      }

    @Test
    public void mustUseTheProvidedDisplayableEvenWhenTheBoundObjectHasGotExpicitDisplayable()
      {
        final Displayable displayable = mock(Displayable.class);
        final Lookup lookup = Lookups.fixed(displayable);
        final Lookup.Provider boundObject = mock(Lookup.Provider.class);
        when(boundObject.getLookup()).thenReturn(lookup);

        final HItem item = fixture.createChild().boundTo(boundObject).withDisplayName("new name").build();
        assertThat(item.as(Displayable), is(not(sameInstance(displayable))));
        assertThat(item.as(Displayable).getDisplayName(), is("new name"));
      }

    @Test(expected=IllegalArgumentException.class)
    public void mustThrowExceptionWhenNoExplicitDisplayNameAndBoundObjectHasNoDisplayName()
      {
        final Lookup.Provider boundObject = mock(Lookup.Provider.class);
        when(boundObject.getLookup()).thenReturn(Lookup.EMPTY);

        fixture.createChild().boundTo(boundObject).build();
      }

    @Test(expected=IllegalArgumentException.class)
    public void mustThrowExceptionWhenNoExplicitDisplayNameAndNoBoundObject()
      {
        fixture.createChild().build();
      }

    @Test
    public void mustMakeTheBoundObjectAvailableAsAPersonality()
      {
        final BoundObject boundObject = new BoundObject();
        final HItem hItem = fixture.createChild().boundTo(boundObject).withDisplayName("name").build();

        assertThat(hItem.as(BoundObject.class), is(sameInstance(boundObject)));
      }

    @Test
    public void mustMakeTheBoundObjectAvailableInTheLookup()
      {
        final BoundObject boundObject = new BoundObject();
        final HItem hItem = fixture.createChild().boundTo(boundObject).withDisplayName("name").build();

        assertThat(hItem.getLookup().lookup(BoundObject.class), is(sameInstance(boundObject)));
      }

    @Test
    public void mustReturnTheChildrenInOrder()
      {
        final HItem item1 = fixture.createChild().withDisplayName("name 1").build();
        final HItem item2 = fixture.createChild().withDisplayName("name 2").build();
        final HItem item3 = fixture.createChild().withDisplayName("name 3").build();

        final List<? extends HItem> children = fixture.findChildren().results();
        assertThat(children, is(notNullValue()));
        assertThat(children.size(), is(3));
        assertThat(children.get(0), is(sameInstance(item1)));
        assertThat(children.get(1), is(sameInstance(item2)));
        assertThat(children.get(2), is(sameInstance(item3)));
      }
  }
