/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: HViewManagerTestSupport.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.hierarchy;

import it.tidalwave.hierarchy.HViewManager;
import it.tidalwave.hierarchy.HView;
import java.util.ArrayList;
import java.util.List;
import it.tidalwave.util.AsException;
import it.tidalwave.util.Id;
import it.tidalwave.util.MutableDisplayable;
import it.tidalwave.util.NotFoundException;
import static it.tidalwave.util.Displayable.Displayable;
import static it.tidalwave.util.MutableDisplayable.MutableDisplayable;
import static it.tidalwave.hierarchy.HItem.Copiable;
import static it.tidalwave.hierarchy.HItem.Movable;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: HViewManagerTestSupport.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
public abstract class HViewManagerTestSupport
  {
    protected HViewManager hvManager;

    @Test
    public void mustHaveTheDisplayName()
      {
        assertThat(hvManager.as(Displayable).getDisplayName(), is("Hierarchic Views"));
      }

    @Test(expected=AsException.class)
    public void mustNotBeAMutableDisplayable()
      {
        hvManager.as(MutableDisplayable);
      }

    //////////////////////// Default View
    @Test
    public void aDefaultViewMustHaveTheCorrectName()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     build();
        assertThat(view, is(notNullValue()));
        assertThat(view.as(Displayable).getDisplayName(), is("My View 1"));
      }

    @Test
    public void aDefaultViewCannotBeRenamed()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     build();
        assertThat(view, is(notNullValue()));
        assertThat(view.as(Displayable), is(not(instanceOf(MutableDisplayable.class))));
      }

    @Test
    public void aDefaultViewMustHaveAnId()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     build();
        assertThat(view, is(notNullValue()));
        assertThat(view.getId(), is(notNullValue()));
      }

    @Test
    public void aDefaultViewMustBeInitiallyEmpty()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     build();
        assertThat(view.findChildren().count(), is(0));
      }

    @Test(expected=AsException.class)
    public void aDefaultViewMustNotBeCopiable()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     build();
        view.as(Copiable);
      }

    @Test(expected=AsException.class)
    public void aDefaultViewMustNotBeMovable()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     build();
        view.as(Movable);
      }

    //////////////////////// Renamable View
    @Test
    public void aRenamableViewMustHaveTheCorrectName()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     thatCanBeRenamed().
                                     build();
        assertThat(view, is(notNullValue()));
        assertThat(view.as(Displayable).getDisplayName(), is("My View 1"));
      }

    @Test
    public void aRenamableViewCanBeRenamed()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     thatCanBeRenamed().
                                     build();
        assertThat(view, is(notNullValue()));
        view.as(MutableDisplayable.class).setDisplayName("New View Name");
        assertThat(view.as(Displayable).getDisplayName(), is("New View Name"));
      }

    @Test
    public void aRenamableViewMustHaveAnId()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     thatCanBeRenamed().
                                     build();
        assertThat(view, is(notNullValue()));
        assertThat(view.getId(), is(notNullValue()));
      }

    @Test
    public void aRenamableViewMustBeInitiallyEmpty()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     thatCanBeRenamed().
                                     build();
        assertThat(view.findChildren().count(), is(0));
      }

    @Test(expected=AsException.class)
    public void aRenamableViewMustNotBeCopiable()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     thatCanBeRenamed().
                                     build();
        view.as(Copiable);
      }

    @Test(expected=AsException.class)
    public void aRenamableViewMustNotBeMovable()
      {
        final HView view = hvManager.createView().
                                     withDisplayName("My View 1").
                                     thatCanBeRenamed().
                                     build();
        view.as(Movable);
      }

    ////// MISC
    @Test
    public void mustCreateViewsWithDifferentIds()
      {
        final List<Id> ids = new ArrayList<Id>();

        for (int i = 0; i < 100; i++)
          {
            final HView view1 = hvManager.createView().
                                          withDisplayName("My View " + i).
                                          build();
            assertFalse(ids.contains(view1.getId()));
            ids.add(view1.getId());
          }
      }

    @Test
    public void mustBeAbleToFindViewById()
      throws NotFoundException
      {
        final HView view1 = hvManager.createView().
                                      withDisplayName("My View 1").
                                      build();
        final HView view2 = hvManager.createView().
                                      withDisplayName("My View 2").
                                      build();

        assertThat(hvManager.findViews().id(view1.getId()).result(), is(sameInstance(view1)));
        assertThat(hvManager.findViews().id(view2.getId()).result(), is(sameInstance(view2)));

        assertThat(hvManager.findViews().id("fakeId").count(), is(0));
      }

    @Test
    public void testFindOrCreateViewWithNoView()
      {
        final HView view = hvManager.findOrCreateView().withId("id1").
                                                        withDisplayName("the name").
                                                        build();
        assertThat(view.getId(), is(new Id("id1")));
      }

    @Test
    public void testFindOrCreateViewWithExistingView()
      {
        final HView view1 = hvManager.createView().withDisplayName("a name").build();
        final HView view2 = hvManager.findOrCreateView().withId(view1.getId()).
                                                         withDisplayName("the name").
                                                         build();
        assertThat(view2, is(sameInstance(view1)));
        assertThat(view1.as(Displayable).getDisplayName(), is("a name"));
      }


    @Test
    public void mustReturnTheViewsInOrder()
      {
        final HView view1 = hvManager.createView().withDisplayName("name 1").build();
        final HView view2 = hvManager.createView().withDisplayName("name 2").build();
        final HView view3 = hvManager.createView().withDisplayName("name 3").build();

        final List<? extends HView> children = hvManager.findViews().results();
        assertThat(children, is(notNullValue()));
        assertThat(children.size(), is(3));
        assertThat(children.get(0), is(sameInstance(view1)));
        assertThat(children.get(1), is(sameInstance(view2)));
        assertThat(children.get(2), is(sameInstance(view3)));
      }
  }