/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: ProgressInputStream.java,v 4e8a394a80a9 2010/03/28 20:51:11 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.downloadprogresspane;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.awt.EventQueue;
import it.tidalwave.util.logging.Logger;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: ProgressInputStream.java,v 4e8a394a80a9 2010/03/28 20:51:11 fabrizio $
 *
 **********************************************************************************************************************/
public class ProgressInputStream extends InputStream
  {
    private final static String CLASS = ProgressInputStream.class.getName();
    private final static Logger logger = Logger.getLogger(CLASS);

    public static final String PROP_TOTAL_BYTE_COUNT = "totalByteCount";
    public static final String PROP_PROCESSED_BYTE_COUNT = "processedByteCount";
    
    @Nonnull
    private final InputStream delegate;

    @Nonnegative
    protected int totalByteCount;

    @Nonnegative
    protected int processedByteCount;

    private final PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);

    /***************************************************************************
     *
     *
     **************************************************************************/
    public ProgressInputStream (final File file)
      throws FileNotFoundException
      {
        this(new FileInputStream(file));
        setTotalByteCountInEDT((int)file.length());
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public ProgressInputStream (@Nonnull final URL url)
      throws IOException
      {
        final URLConnection connection = url.openConnection();
        final int length = connection.getContentLength();
        final String mimeType = connection.getContentType();
        logger.finer(">>>> length from HTTP headers: %d", length);
        logger.finer(">>>> MIME type from HTTP headers: %s", mimeType);

        if (length > 0)
          {
            setTotalByteCountInEDT(length);
          }

        this.delegate = connection.getInputStream();
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public ProgressInputStream (@Nonnull final InputStream delegate,
                                @Nonnegative final int size)
      {
        this.delegate = delegate;
        setTotalByteCountInEDT(size);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public ProgressInputStream (@Nonnull final InputStream delegate)
      {
        this.delegate = delegate;
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public long skip (long n)
      throws IOException
      {
        incrementProcessedByteCountInEDT((int)n);
        return delegate.skip(n);
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public synchronized void reset() 
      throws IOException
      {
        delegate.reset();
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public int read (byte[] buffer, int offset, int length)
      throws IOException
      {
        final int n = delegate.read(buffer, offset, length);
        incrementProcessedByteCountInEDT(n);
        return n;
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public int read (byte[] arg0)
      throws IOException
      {
        final int n = delegate.read(arg0);
        incrementProcessedByteCountInEDT(n);
        return n;
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public int read()
      throws IOException
      {
        final int n = delegate.read();
        incrementProcessedByteCountInEDT((n >= 0) ? 1 : 0);
        return n;
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public boolean markSupported()
      {
        return delegate.markSupported();
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public synchronized void mark (int arg0)
      {
        delegate.mark(arg0);
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public void close() 
      throws IOException
      {
        delegate.close();
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public int available() 
      throws IOException
      {
        return delegate.available();
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    protected void setProcessedByteCount (@Nonnegative final int processedByteCount)
      {
        int oldProcessedByteCount = this.processedByteCount;
        this.processedByteCount = processedByteCount;
        propertyChangeSupport.firePropertyChange(PROP_PROCESSED_BYTE_COUNT, oldProcessedByteCount, processedByteCount);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnegative
    public int getProcessedByteCount()
      {
        return processedByteCount;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    protected void setTotalByteCount (@Nonnegative final int totalByteCount)
      {
        int oldTotalByteCount = this.totalByteCount;
        this.totalByteCount = totalByteCount;
        propertyChangeSupport.firePropertyChange(PROP_TOTAL_BYTE_COUNT, oldTotalByteCount, totalByteCount);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public int getTotalByteCount()
      {
        return totalByteCount;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public void addPropertyChangeListener (PropertyChangeListener listener)
      {
        propertyChangeSupport.addPropertyChangeListener(listener);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public void removePropertyChangeListener (PropertyChangeListener listener)
      {
        propertyChangeSupport.removePropertyChangeListener(listener);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    private void setTotalByteCountInEDT (final int totalByteCount)
      {
        EventQueue.invokeLater(new Runnable()
          {
            public void run()
              {
                setTotalByteCount(totalByteCount);
              }
          });
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    private void incrementProcessedByteCountInEDT (final int increment)
      {
        EventQueue.invokeLater(new Runnable()
          {
            public void run()
              {
                setProcessedByteCount(getProcessedByteCount() + increment);
              }
          });
      }
  }
