/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: IndexedFileSystem.java,v 9607f67dd36d 2010/04/04 22:40:40 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.indexedfilesystem;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.Collection;
import java.io.File;
import java.io.Serializable;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.LocalFileSystem;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.indexedfilesystem.spi.FileIndexer;

/***********************************************************************************************************************
 *
 * The <code>IndexedFileSystem</code> is a specialization of
 * <code>LocalFileSystem</code> that binds an unique id to any <code>FileObject</code>.
 * All instances of <code>FileObject</code> returned by methods in this class
 * have the special attribute <code>ID_ATTRIBUTE</code> which contains their id.
 *
 * @author  Fabrizio Giudici
 * @version $Id: IndexedFileSystem.java,v 9607f67dd36d 2010/04/04 22:40:40 fabrizio $
 *
 **********************************************************************************************************************/
public class IndexedFileSystem extends LocalFileSystem
  {
    private static final String CLASS = IndexedFileSystem.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    private static final long serialVersionUID = 6509911115994045263L;
    
    /** The <code>FileObject</code> attribute which contains the unique id. */
    public static final String ID_ATTRIBUTE = "it.tidalwave.netbeans.indexedfilesystem.id";

    private FileIndexer fileIndexer = Locator.find(FileIndexer.class);

    /***************************************************************************
     *
     * Creates a new instance.
     *
     **************************************************************************/
    public IndexedFileSystem()
      {
        try 
          {
//            setRootDirectory(new File("C:\\")); // FIXME: Windows?
          } 
        catch (Exception e) 
          {
            throw new RuntimeException(e);
          } 
      } 
    
        /***************************************************************************
         *
         * @inheritDoc
         *
         **************************************************************************/
//        @Override
//        void beginAtomicAction (FileSystem.AtomicAction run) 
//          {
//            // TODO: consider beginning a transaction
//            super.beginAtomicAction(run);
//          }

        /***************************************************************************
         *
         * @inheritDoc
         *
         **************************************************************************/
//        @Override
//        void finishAtomicAction() 
//          {
//            super.finishAtomicAction();
//            // TODO: consider committing a transaction
//          }
//      }

    /***************************************************************************
     * 
     * @inheritDoc
     *
     **************************************************************************/
    @Override
    @Nonnull
    public FileObject findResource (final @Nonnull String path)
      {
        return fileIndexer.findResource(path);
      }
    
    /***************************************************************************
     * 
     * Returns true if the given path matches an indexed resource. This method
     * is mostly used for testing.
     * 
     * @param  path  a path
     * @return       true if the path is indexed
     *
     **************************************************************************/
    public boolean isIndexed (final @Nonnull String path)
      {
        return fileIndexer.isIndexed(path);    
      }
    
    /***************************************************************************
     * 
     * @inheritDoc
     *
     **************************************************************************/
    @Override
    @Nonnull
    public String getDisplayName()
      {
        return "My Computer"; // FIXME
      }
    
    /***************************************************************************
     *
     * Returns a collection of the current file roots.
     * 
     * @return  the file roots
     *
     **************************************************************************/
    @Nonnull
    public Collection<FileObject> findRoots()
      {
        final Collection<FileObject> result = new ArrayList<FileObject>();

        for (final File root : File.listRoots())
          {
            try
              {
                result.add(findResource(root.getAbsolutePath()));
              }
            catch (Exception e)
              {
                logger.warning("Ignoring %s because of %s", root, e);
              }
          }

        logger.info("getRoots() returning %s", result);
        return result;
      }
    
    /***************************************************************************
     * 
     * Finds the path bound to the given id. It is first searched in the local
     * cache, then in the persistent storage. In this case the pending mappings
     * are flushed.
     *
     * @param   id   the id to search
     * @return       the FileObject bound to the given id
     *
     **************************************************************************/
    @Nonnull
    public FileObject findResourceById (final @Nonnull Serializable id)
      {
        return fileIndexer.findResourceById(id);
      }

    /***************************************************************************
     * 
     * Returns a collection of all the indexed ids.
     * 
     * @return  all the indexed ids
     *
     **************************************************************************/
    @Nonnull
    public Collection<Serializable> findAllIndexedIds() 
      {
        return fileIndexer.findAllIndexedIds();
      }

    /***************************************************************************
     * 
     * Removes the indexing of the given FileObject, without deleting the file.
     * If there's no indexing for the given FileObject, the method silently
     * returns.
     *
     * @param   fileObject  the file whose indexing must be removed
     *
     **************************************************************************/
    public void removeIndexing (final @Nonnull FileObject fileObject)
      {
        fileIndexer.removeIndexing(fileObject);
      }

    /***************************************************************************
     * 
     * Removes the indexing of the given id, without deleting the file.
     * If there's no indexing for the given FileObject, the method silently
     * returns.
     *
     * @param   id  the id whose indexing must be removed
     *
     **************************************************************************/
    public void removeIndexing (final @Nonnull Serializable id)
      {
        fileIndexer.removeIndexing(id);
      }
  } 
