/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: AttrDecorator.java,v ab5e96a57544 2010/04/04 20:12:06 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.indexedfilesystem.impl;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Enumeration;
import java.io.IOException;
import java.io.Serializable;
import org.openide.filesystems.AbstractFileSystem;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.indexedfilesystem.IndexedFileSystem;
import it.tidalwave.netbeans.indexedfilesystem.spi.FileIndexer;

/***********************************************************************************************************************
 *
 * This class intercepts operations on attributes and provides the semantics for the ID_ATTRIBUTE.
 *
 * @author  Fabrizio Giudici
 * @version $Id: AttrDecorator.java,v ab5e96a57544 2010/04/04 20:12:06 fabrizio $
 *
 **********************************************************************************************************************/
class AttrDecorator implements AbstractFileSystem.Attr
  {
    private static final String CLASS = AttrDecorator.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    private static final long serialVersionUID = -7402108821245557936L;
    
    private static final FileIndexer FILE_INDEXER = Locator.find(FileIndexer.class);

    @Nonnull
    private final AbstractFileSystem.Attr delegate;
    
    @Nonnull
    private final LocalIndexedFileSystem fileSystem;
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public AttrDecorator (final @Nonnull AbstractFileSystem.Attr delegate,
                          final @Nonnull LocalIndexedFileSystem fileSystem)
      {
        this.delegate = delegate;
        this.fileSystem = fileSystem;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nullable
    public Object readAttribute (final @Nonnull String path, final @Nonnull String name)
      {
//        logger.fine("readAttribute(" + path + " ," + name + ")");
        
        if (IndexedFileSystem.ID_ATTRIBUTE.equals(name))
          {
            logger.fine("readAttribute(%s, %s)", path, name);
            final String path2 = fileSystem.findCompletePath(path);
            logger.finer(">>> fixed path: %s", path2);
            Serializable id = FILE_INDEXER.findIdByPath(path2);

            if (id == null)
              {
                FILE_INDEXER.createIndexing(path2, false);
                id = FILE_INDEXER.findIdByPath(path2);
              }

            return id;
          }

        else
          {
            return delegate.readAttribute(path, name);
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void writeAttribute (final @Nonnull String path,
                                final @Nonnull String name,
                                final @Nonnull Object value)
      throws IOException
      {
        if (IndexedFileSystem.ID_ATTRIBUTE.equals(name))
          {
            throw new IllegalArgumentException("Attribute " + name + " is read-only");
          }

        delegate.writeAttribute(path, name, value);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Enumeration attributes (final @Nonnull String path)
      {
        final Enumeration e = delegate.attributes(path);

        return new Enumeration() 
          {
            private boolean first = true;

            @Override
            public boolean hasMoreElements() 
              {
                return first || e.hasMoreElements();
              }

            @Override
            public Object nextElement() 
              {
                Object o = first ? IndexedFileSystem.ID_ATTRIBUTE : e.nextElement();
                first = false;
                return o;
              }
          };
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void renameAttributes (final @Nonnull String oldPath, final @Nonnull String newPath)
      {
        logger.fine("renameAttributes(%s, %s)", oldPath, newPath);
        delegate.renameAttributes(oldPath, newPath);
        final String newPath2 = fileSystem.findCompletePath(newPath);
        final String oldPath2 = fileSystem.findCompletePath(oldPath);
        logger.finer(">>> fixed paths: %s, %s", oldPath2, newPath2);

        if (FILE_INDEXER.isIndexed(oldPath2))
          {
            FILE_INDEXER.handleFileRenamed(oldPath2, newPath2);
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void deleteAttributes (final @Nonnull String path)
      {
        logger.fine("deleteAttributes(%s)", path);
        delegate.deleteAttributes(path);
        final String path2 = fileSystem.findCompletePath(path);
        logger.finer(">>> fixed path: %s", path2);

        if (FILE_INDEXER.isIndexed(path2))
          {
            FILE_INDEXER.handleFileDeleted(path2);
          }
      }
  }
