/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: IndexedFileEntity.java,v ab5e96a57544 2010/04/04 20:12:06 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.indexedfilesystem.jpa;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.util.Date;
import java.io.Serializable;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.Id;
import javax.persistence.Table;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.persistence.Persistence;

/***********************************************************************************************************************
 *
 * This class is used for persistent mapping (id;path).
 *
 * @author  Fabrizio Giudici
 * @version $Id: IndexedFileEntity.java,v ab5e96a57544 2010/04/04 20:12:06 fabrizio $
 *
 **********************************************************************************************************************/
@Entity
@Table(name="INDEXED_FILE")
public class IndexedFileEntity implements Serializable
  {
    private static final String CLASS = IndexedFileEntity.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    
    private static int instanceCount = 0;
    private static int creationCount = 0;
    
    @Id
    @Column(name="ID",length=40)
    // Id is explicitly assigned by the constructor
    private String id;
    
    @Column(name="PATH",length=760,unique=true,nullable=false)
    private String path;
    
    @Column(name="DIGEST")
    private String digest;
    
    @Column(name="INDEXING_TIMESTAMP")
    private long indexingTimestamp;
    
    @Column(name="DIGEST_TIMESTAMP")
    private Long digestTimestamp;
    
    /***************************************************************************
     *
     * For JPA only.
     *
     **************************************************************************/
    protected IndexedFileEntity()
      {
        count(+1, +1);
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public IndexedFileEntity (@Nonnull final String path)
      {
        this();
        
        this.id = Locator.find(Persistence.class).createUniqueId().toString();
        setPath(path);
        this.indexingTimestamp = System.currentTimeMillis();
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public String getId()
      {
        return id;  
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public String getPath()
      {
        return path;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public void setPath (@Nonnull final String path)
      {
        if (path == null)
          {
            throw new IllegalArgumentException("path is mandatory");
          }
        
        if (path.indexOf('\\') >= 0)
          {
            throw new IllegalArgumentException("path cannot contain back slashes");
          }
        
        this.path = path;
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    @CheckForNull
    public String getDigest() 
      {
        return digest;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public void setDigest (@Nonnull final String digest)
      {
        this.digest = digest;
        digestTimestamp = System.currentTimeMillis();
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public Date getIndexingTimestamp() 
      {
        return new Date(indexingTimestamp);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @CheckForNull
    public Date getDigestTimestamp() 
      {
        return (digestTimestamp == null) ? null : new Date(digestTimestamp);
      }  

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public int hashCode()
      {
        return (id != null) ? id.hashCode() : 0;
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public boolean equals (@CheckForNull final Object object)
      {
        if (!(object instanceof IndexedFileEntity))
          {
            return false;      
          }
        
        final IndexedFileEntity other = (IndexedFileEntity)object;
        
        if (id == null)
          {
            return false;  
          }
        
        return this.id.equals(other.id);
      }
    
    /***************************************************************************
     *
     * @inheritDoc
     *
     **************************************************************************/
    @Override @Nonnull
    public String toString()
      {
        return String.format("IndexedFileEntity[id:%s, path:%s, digest:%s]", id, path, digest);  
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    @Override
    protected void finalize()
      {
        count(-1, 0);
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    private final static void count (final int instanceDelta, final int creationDelta)
      {
        instanceCount += instanceDelta;
        creationCount += creationDelta;
        logger.finest("STATS: ########### IndexedFileEntity instances: %s, created so far: %d ", instanceCount, creationCount);
      }
  }
