/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: JPAFileIndexer.java,v b79f5552e438 2010/04/04 23:08:30 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.indexedfilesystem.jpa;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Collection;
import java.util.List;
import java.io.Serializable;
import org.openide.util.lookup.ServiceProvider;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.persistence.TxTask;
import it.tidalwave.netbeans.persistence.Persistence;
import it.tidalwave.netbeans.indexedfilesystem.spi.FileIndexer;
import it.tidalwave.netbeans.indexedfilesystem.spi.FileIndexerSupport;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: JPAFileIndexer.java,v b79f5552e438 2010/04/04 23:08:30 fabrizio $
 *
 **********************************************************************************************************************/
@ServiceProvider(service=FileIndexer.class)
public class JPAFileIndexer extends FileIndexerSupport
  {
    private static final String CLASS = JPAFileIndexer.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    
    /***************************************************************************
     * 
     *
     **************************************************************************/
    public JPAFileIndexer()
      {
        initializeDatabase();
      }
    
    /***************************************************************************
     * 
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override @Nonnull
    public synchronized Collection<Serializable> findAllIndexedIds()
      {
        return TxTask.execute(new TxTask<Collection<Serializable>, RuntimeException>() 
          {
            @Override
            public Collection<Serializable> run()  
              {
                queryCount++;
                return query("select i.id from IndexedFileEntity i");
              }
          });
      }
    
    /***************************************************************************
     * 
     * Finds the path bound to the given id. It is first searched in the local
     * cache, then in the persistent storage. In this case the pending mappings
     * are flushed.
     *
     * @param   id   the id to search
     * @return       the path bound to the given id
     *
     **************************************************************************/
    @Override @Nullable
    public synchronized String findPathById (final @Nonnull Serializable id)
      {
        if (!normalizedPathMapById.containsKey(id) && !normalizedPathPendingSet.isEmpty())
          {
            flushPending();  
          }
        
        String normalizedPath = normalizedPathMapById.get(id);
        
        if (normalizedPath == null)
          {
            normalizedPath = TxTask.execute(new TxTask<String, RuntimeException>() 
              {
                @Override
                public String run()
                  {
                    queryCount++;
                    final IndexedFileEntity indexedFileEntity = getEntityManager().find(IndexedFileEntity.class, id);
                    return (indexedFileEntity != null) ? indexedFileEntity.getPath() : null;
                  }
              });
            
            if (normalizedPath != null) // defensive
              {
                normalizedPathMapById.put(id, normalizedPath);
              }
          }
        
        if (normalizedPath == null) 
          {
            logger.warning("No IndexedFileObject for id = %s", id);
            return null;
          }
        
        return localizedPath(normalizedPath);
      }
            
    /***************************************************************************
     * 
     * Handles a file deletion in the underlying filesystem. The file is removed
     * from both the memory caches and the persistent store.
     *
     * @param  path  the deleted file
     *
     **************************************************************************/
    @Override
    public void handleFileDeleted (final @Nonnull String path)
      {
        logger.fine("handleFileDeleted(%s)", path);
        
        final String normalizedPath = normalizedPath(path);
        normalizedPathPendingSet.remove(normalizedPath);
        normalizedPathWriteSet.remove(normalizedPath);
        Serializable id = idMapByNormalizedPath.remove(normalizedPath);
        assert id != null : "null id for path: " + path;
        normalizedPathMapById.remove(id);

        TxTask.execute(new TxTask<Void, RuntimeException>()
          {
            @Override
            public Void run()
              {
                try
                  {
                    queryCount++;
                    final IndexedFileEntity indexedFileEntity = querySingle("select i from IndexedFileEntity i where i.path = :path", "path", normalizedPath);
                    getEntityManager().remove(indexedFileEntity);
                  }
                catch (NotFoundException e)
                  {
                    // ok, sometimes the id has not been flushed yet so there's no entity
                  }
                
                return null;
              }
          });
      }

    /***************************************************************************
     * 
     * Updates a file mapping.
     *
     * @param  newNormalizedPath  the new path
     * @param  oldNormalizedPath  the old path
     *
     **************************************************************************/
    @Override
    protected void updateMapping (final @Nonnull String newNormalizedPath, final @Nonnull String oldNormalizedPath)
      {
        logger.finer("updateMapping(%s -> %s)", oldNormalizedPath, newNormalizedPath);
        
        normalizedPathPendingSet.remove(oldNormalizedPath);
        normalizedPathWriteSet.remove(oldNormalizedPath);
        final Serializable id = idMapByNormalizedPath.remove(oldNormalizedPath);
        assert id != null : "null id";
        normalizedPathMapById.remove(id);

        TxTask.execute(new TxTask<Void, RuntimeException>()
          {
            @Override
            public Void run() 
              {
                try
                  {
                    queryCount++;
                    final IndexedFileEntity indexedFileEntity = querySingle("select i from IndexedFileEntity i where i.path = :path", "path", oldNormalizedPath);
                    indexedFileEntity.setPath(newNormalizedPath);
                    final Serializable id = indexedFileEntity.getId();
                    idMapByNormalizedPath.put(newNormalizedPath, id);
                    normalizedPathMapById.put(id, newNormalizedPath);
                  }
                catch (NotFoundException e)
                  {
                    // ok, sometimes the id has not been flushed yet so there's no entity
                    // FIXME: is the above right?
                  }

                return null;
              }
          });
      }
    
    /***************************************************************************
     * 
     * Flushes the mapping requests in the pending list. Existing mappings are 
     * loaded into the cache, new mappings are stored to the persistent layer.
     *
     **************************************************************************/
    @Override
    public synchronized void flushPending()
      {
        if (normalizedPathPendingSet.isEmpty()) // might happen in some unit tests
          {
            return;  
          }
        
        TxTask.execute(new TxTask<Void, RuntimeException>()
          {
            @Override
            public Void run()
              {
                logger.fine("Flushing pending list (%d items)", normalizedPathPendingSet.size());
                queryCount++;
                // TODO: replace with persistence.query(...) - but check first query.setCacheable()!
//                Query query = session.createQuery("from IndexedFilePB where path in (:paths)");
//                query.setParameterList("paths", normalizedPathPendingSet);
//                query.setCacheable(true);
//                List result = query.list();
                // FIXME: we are not setting setCacheable(true)
                final List<IndexedFileEntity> result = query("select i from IndexedFileEntity i where i.path in (:paths)", "paths", normalizedPathPendingSet);
                logger.finer(">>>> returned: %d objects",  result.size());
//                logger.finest(">>>> returned: " + result);
                //
                // Load the existing bindings into the cache.
                //
                for (final IndexedFileEntity indexedFileEntity : result)
                  {
                    final String normalizedPath = indexedFileEntity.getPath();
                    final Serializable id = indexedFileEntity.getId();
                    logger.finer(">>>> Retrieved id: %s for path: %s", id, normalizedPath);
                    idMapByNormalizedPath.put(normalizedPath, id);
                    normalizedPathMapById.put(id, normalizedPath);
                    normalizedPathPendingSet.remove(normalizedPath);
                    normalizedPathWriteSet.remove(normalizedPath);
                  }
                //
                // Store the new bindings that were only in the pending list.
                //
                for (final String normalizedPath : normalizedPathWriteSet)
                  {
                    final IndexedFileEntity indexedFileEntity = new IndexedFileEntity(normalizedPath);
                    final Serializable id = indexedFileEntity.getId();
                    getEntityManager().persist(indexedFileEntity);
//                    session.persist(indexedFilePB);
//                    session.flush();
//                    session.refresh(indexedFilePB);
                    idMapByNormalizedPath.put(normalizedPath, id);
                    normalizedPathMapById.put(id, normalizedPath);
                    logger.finer(">>>> Created new id: %s for path: %s", id, normalizedPath);
                  }
                
                normalizedPathPendingSet.clear();   
                normalizedPathWriteSet.clear();
                return null;
              }
          });
      }
    
    /***************************************************************************
     *
     **************************************************************************/
    private synchronized void initializeDatabase()
      {
        TxTask.execute(new TxTask<Void, RuntimeException>() 
          {
            @Override
            public Void run()
              {
                final Persistence persistence = Locator.find(Persistence.class);
                persistence.createIndexes(IndexedFileEntity.class, "path", "digest", "indexingTimestamp", "digestTimestamp");
                return null;
              }
          });
      }
  }
