/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: FileIndexer.java,v 04aade4dda6b 2010/04/04 21:17:46 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.indexedfilesystem.spi;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Collection;
import java.io.Serializable;
import org.openide.filesystems.FileObject;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: FileIndexer.java,v 04aade4dda6b 2010/04/04 21:17:46 fabrizio $
 *
 **********************************************************************************************************************/
public interface FileIndexer
  {
    /*******************************************************************************************************************
     *
     * 
     *
     ******************************************************************************************************************/
    public void removeIndexing (@Nonnull FileObject fileObject);

    /*******************************************************************************************************************
     *
     *
     *
     ******************************************************************************************************************/
    public void removeIndexing (@Nonnull Serializable id);

    /*******************************************************************************************************************
     *
     * Returns true if the given path is indexed (or it has been scheduled for indexing). This method doesn't try to
     * create a new indexing, as it is used for tessting an existing indexing. This method is mainly for testing and
     * this is the reason for which it accepts a path as a parameter instead of a FileObject (it is impossible to create
     * a FileObject instance for a not existing file).
     *
     * @param   path  the path
     *
     ******************************************************************************************************************/
    public boolean isIndexed (@Nonnull String path);

    /*******************************************************************************************************************
     *
     * 
     *
     ******************************************************************************************************************/
    @Nullable
    public FileObject findResourceById (@Nonnull Serializable id);

    /*******************************************************************************************************************
     *
     * Finds the id bound to the given path. It is first searched in the local cache, then in the persistent storage. In
     * this case the pending mappings are flushed.
     *
     * @param  path  the path to map
     * @return       the id bound to the path
     *
     ******************************************************************************************************************/
    @Nullable
    public Serializable findIdByPath (@Nonnull String path);

    /*******************************************************************************************************************
     *
     *
     *
     ******************************************************************************************************************/
    @CheckForNull
    public FileObject findResource (@Nonnull String path);

    /*******************************************************************************************************************
     *
     * Creates an indexing for the given path. Indeed the indexing is not created immediately: it will be done when
     * the id attribute is requested for the first time. Indexed files are accumulated in a buffer so a batch operation
     * will be performed: <code>flushPending()</code> will both load all the existing ids and create the missing ones.
     * If prefetch-only mode is specified, id creation won't be performed in batch mode.
     *
     * @param   path          the path of file to index
     * @param   prefetchOnly  true if prefetch only
     * @return                the indexed file
     *
     ******************************************************************************************************************/
    public void createIndexing (@Nonnull String path, boolean prefetchOnly);

    /*******************************************************************************************************************
     *
     * Normalizes the path. A normalized path has always forward slashes as file separators; furthermore on the Windows
     * platform, the leading C: is stripped.
     *
     * @param  path  the path to normalize
     * @return       the normalized path
     *
     ******************************************************************************************************************/
    @Nonnull
    public String normalizedPath (@Nonnull String path);

    /*******************************************************************************************************************
     *
     * Localizes a path. This operation is the opposite of normalizedPath().
     *
     * @param  path  the path to localize
     * @param        the localized path
     *
     ******************************************************************************************************************/
    @Nonnull
    public String localizedPath (@Nonnull String path);

    /*******************************************************************************************************************
     *
     * Flushes all the internal caches. This method is only for unit testing;
     * in complex scenarios can lead to data corruption.
     *
     ******************************************************************************************************************/
    public void flushCaches();

    /*******************************************************************************************************************
     *
     * Handles a file rename in the underlying filesystem. The file is renamed
     * both in the memory caches and the persistent store; if the file is a
     * directory, a recursive rename is applied to all contents.
     *
     * @param  oldPath  the old path of the file object
     * @param  newPath  the new path of the file object
     *
     ******************************************************************************************************************/
    public void handleFileRenamed (@Nonnull String oldPath, @Nonnull String newPath);

    /*******************************************************************************************************************
     *
     * Rename recursively a folder.
     *
     *
     * @param fileObject            the file to rename
     * @param oldNormalizedPath     the old path
     * @param newNormalizedPath     the new path
     *
     ******************************************************************************************************************/
    public void renameRecursively (@Nonnull FileObject fileObject,
                                   @Nonnull String oldNormalizedPath,
                                   @Nonnull String newNormalizedPath);

    /*******************************************************************************************************************
     *
     *
     *
     ******************************************************************************************************************/
    public void handleFileDeleted (@Nonnull String path);

    /*******************************************************************************************************************
     *
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Collection<Serializable> findAllIndexedIds();

    /*******************************************************************************************************************
     *
     * Finds the path bound to the given id. It is first searched in the local cache, then in the persistent storage. In
     * this case the pending mappings are flushed.
     *
     * @param   id   the id to search
     * @return       the path bound to the given id
     *
     ******************************************************************************************************************/
    @Nonnull
    public String findPathById (@Nonnull Serializable id);

    /*******************************************************************************************************************
     *
     *
     *
     ******************************************************************************************************************/
    public void flushPending();
  }
