/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: FileIndexerSupport.java,v 04aade4dda6b 2010/04/04 21:17:46 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.indexedfilesystem.spi;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.beans.PropertyVetoException;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import org.openide.filesystems.FileObject;
import org.openide.util.Utilities;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.indexedfilesystem.impl.LocalIndexedFileSystem;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: FileIndexerSupport.java,v 04aade4dda6b 2010/04/04 21:17:46 fabrizio $
 *
 **********************************************************************************************************************/
public abstract class FileIndexerSupport implements FileIndexer
  {
    private static final String CLASS = FileIndexerSupport.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    private static final String C_DRIVE = "C:";

    /** The maximum number of files to update in batch mode. */
    private static final int MAX_PENDING_COUNT = 50;

    /** A cache of local file systems indexer by their root. */
    protected final Map<String, LocalIndexedFileSystem> fileSystemMapByRoot = new HashMap<String, LocalIndexedFileSystem>();

    /** List of paths of indexed files that have been created but not asked for the id yet. */
    protected final Set<String> normalizedPathPendingSet = new HashSet<String>();

    /** List of paths of indexed files that need an id creation, if doesn't exist, at the next flush. */
    protected final Set<String> normalizedPathWriteSet = new HashSet<String>();

    /** A map from paths to id. Paths here are normalized. */
    protected final Map<String, Serializable> idMapByNormalizedPath = new HashMap<String, Serializable>(); // TODO: use weak refs etc...

    /** A map from ids to path. Paths here are normalized. */
    protected final Map<Serializable, String> normalizedPathMapById = new HashMap<Serializable, String>(); // TODO: use weak refs etc...

    /** The number of performed queries. */
    public /* @TestOnly */ int queryCount; // FIXME: the public

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void removeIndexing (final @Nonnull FileObject fileObject)
      {
        if (fileObject == null)
          {
            throw new IllegalArgumentException("fileObject is mandatory");
          }

        handleFileDeleted(fileObject.getPath());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void removeIndexing (final @Nonnull Serializable id)
      {
        if (id == null)
          {
            throw new IllegalArgumentException("id is mandatory");
          }

        final String path = findPathById(id);

        if (path != null)
          {
            // FIXME: maybe this could be moved into handleFileDeleted()?
            createIndexing(path, true);
            flushPending();
            // END FIXME
            handleFileDeleted(path);
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean isIndexed (final @Nonnull String path)
      {
        if (path == null)
          {
            throw new IllegalArgumentException("path is mandatory");
          }

        return findIdByPath(path) != null;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nullable
    public synchronized FileObject findResourceById (final @Nonnull Serializable id)
      {
        final String path = findPathById(id);

        if (path == null)
          {
            return null;
          }

        else
          {
            final File file = new File(path);

            if (!file.exists()) // FileObject requires the actual file to exist
              {
                logger.warning("File does not exist: %s", file);
                return null;
              }
            else
              {
                return findResource(path);
              }
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nullable
    public synchronized Serializable findIdByPath (final @Nonnull String path)
      {
        final String normalizedPath = normalizedPath(path);

        if (!idMapByNormalizedPath.containsKey(normalizedPath) && !normalizedPathPendingSet.isEmpty())
          {
            flushPending();
          }

        return idMapByNormalizedPath.get(normalizedPath);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @CheckForNull
    public synchronized FileObject findResource (final @Nonnull String path)
      {
        final StringTokenizer st = new StringTokenizer(path, "\\/");
        String root = null;

        if (Utilities.isWindows())
          {
            root = st.nextToken() + "\\";
            assert root.length() <= 3;
          }

        else
          {
            root = "/";
          }

        LocalIndexedFileSystem fileSystem = fileSystemMapByRoot.get(root);

        if (fileSystem == null)
          {
            fileSystem = new LocalIndexedFileSystem();

            try
              {
                fileSystem.setRootDirectory(new File(root));
              }
            catch (IOException e)
              {
                throw new RuntimeException(e);
              }
            catch (PropertyVetoException e)
              {
                throw new RuntimeException(e);
              }

            fileSystemMapByRoot.put(root, fileSystem);
          }

        FileObject fo = fileSystem.getRoot();

        // TODO: maybe could be replaced by return fileSystem.findResource();

        //
        // fo.getFileObject() will call ListDecorator.children(), which prefetches
        // indexing for all children. We don't want to prefetch indexes now since
        // the directory tree traversal is done only to get to the desired file.
        // We are running in a critical section as setting a flag in fileSystem is
        // not thread-safe. A better approach could be to set/unset a special
        // volatile attribute on fo. Or a ThreadLocal attribute to the thread.
        //
        synchronized (fileSystem.DISABLE_CHILDREN_INDEXING_LOCK)
          {
            fileSystem.disableChildrenIndexing = true;

            try
              {
                while (st.hasMoreTokens() && (fo != null))
                  {
                    final String s = st.nextToken();
                    final int i = s.lastIndexOf('.');
                    final String base = (i < 0) ? s : s.substring(0, i);
                    final String ext = (i < 0) ? "" : s.substring(i + 1);
                    FileObject child = fo.getFileObject(base, ext);

                    if (child == null)
                      {
                        fo.refresh();
                        child = fo.getFileObject(base, ext);
                      }

                    fo = child;
                  }
              }
            finally
              {
                fileSystem.disableChildrenIndexing = false;
              }
          }

        return fo;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public synchronized void createIndexing (final @Nonnull String path, final boolean prefetchOnly)
      {
        logger.finer("createIndexing(%s, prefetchOnly=%s)", path, prefetchOnly);

        if (path == null)
          {
            throw new IllegalArgumentException("path is mandatory");
          }

        final String normalizedPath = normalizedPath(path);
        final Serializable id = idMapByNormalizedPath.get(normalizedPath);

        if (id == null)
          {
            normalizedPathPendingSet.add(normalizedPath);

            if (!prefetchOnly)
              {
                normalizedPathWriteSet.add(normalizedPath);
              }

            logger.finest(">>>> normalizedPathPendingSet size: %s normalizedPathWriteSet size: %d",
                    normalizedPathPendingSet.size(),
                    normalizedPathWriteSet.size());

            if (normalizedPathPendingSet.size() >= MAX_PENDING_COUNT)
              {
                flushPending();
              }
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String normalizedPath (@Nonnull String path)
      {
        if (Utilities.isWindows())
          {
            path = path.replace('\\', '/');

            if (path.toUpperCase().startsWith(C_DRIVE))
              {
                path = path.substring(C_DRIVE.length());
              }
          }

        boolean isWindowRoot = Utilities.isWindows() && (path.length() >= 2) && (path.charAt(1) == ':');

        if (!path.startsWith("/") && !isWindowRoot) // FileObject paths don't start with /
          {
            path = "/" + path;
          }

        if (path.endsWith("/") && (path.length() > 1))
          {
            path = path.substring(0, path.length() - 1);
          }

        return path;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String localizedPath (@Nonnull String path)
      {
        if (Utilities.isWindows())
          {
            path = path.replace('/', '\\');

            if (path.startsWith("\\"))
              {
                path = C_DRIVE + path;
              }
          }

        return path;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void flushCaches()
      {
        normalizedPathMapById.clear();
        idMapByNormalizedPath.clear();
        normalizedPathPendingSet.clear();
        normalizedPathWriteSet.clear();
        fileSystemMapByRoot.clear();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void handleFileRenamed (final @Nonnull String oldPath, final @Nonnull String newPath)
      {
        // TODO: make it transactional
        final String newNormalizedPath = normalizedPath(newPath);
        final String oldNormalizedPath = normalizedPath(oldPath);
        updateMapping(newNormalizedPath, oldNormalizedPath);
//        FileObject fileObject = findResource(newPath);
//        renameRecursively(fileObject, oldNormalizedPath, newNormalizedPath);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void renameRecursively (final @Nonnull FileObject fileObject,
                                   final @Nonnull String oldNormalizedPath,
                                   final @Nonnull String newNormalizedPath)
      {
        logger.finer("renameRecursively(%s: %s -> %s)", fileObject.getPath(), oldNormalizedPath, newNormalizedPath);

        for (final FileObject child : fileObject.getChildren())
          {
            final String newChildNormalizedPath = normalizedPath(newNormalizedPath + "/" + child.getNameExt());
            final String oldChildNormalizedPath = normalizedPath(oldNormalizedPath + "/" + child.getNameExt());
            updateMapping(newChildNormalizedPath, oldChildNormalizedPath);
            renameRecursively(child, oldNormalizedPath, newNormalizedPath);
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public abstract void handleFileDeleted (@Nonnull String path);

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public abstract Collection<Serializable> findAllIndexedIds();

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public abstract String findPathById (@Nonnull Serializable id);

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public abstract void flushPending();

    /*******************************************************************************************************************
     *
     *
     *
     ******************************************************************************************************************/
    protected abstract void updateMapping (@Nonnull String newNormalizedPath, @Nonnull String oldNormalizedPath);
  }
