/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: TestSupport.java,v d3bc6735d000 2010/04/04 23:40:46 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.indexedfilesystem;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.util.Utilities;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.persistence.Persistence;
import it.tidalwave.netbeans.util.test.TestLoggerSetup;
import javax.annotation.Nonnegative;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import static org.junit.Assert.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: TestSupport.java,v d3bc6735d000 2010/04/04 23:40:46 fabrizio $
 *
 **********************************************************************************************************************/
public abstract class TestSupport
  {
    protected static final Logger logger = Logger.getLogger(TestSupport.class.getName());
    protected static final int FOLDER_COUNT = 10;
    protected static final List<File> testFiles = new ArrayList<File>();
    protected static final List<File> baseFolders = new ArrayList<File>();
    private static int rootsTested;
    protected static final File TEMP_FOLDER = new File("target/TempFolder").getAbsoluteFile();
    protected static final File WORKSPACE = new File(TEMP_FOLDER, "TestWorkspace");
    
    static
      {
        TEMP_FOLDER.mkdirs();
        System.setProperty("it.tidalwave.netbeans.loggerconfiguration.debug", "true");
        System.setProperty("it.tidalwave.netbeans.workspacemanager.workspace", WORKSPACE.getAbsolutePath());
        System.setProperty("it.tidalwave.netbeans.workspacemanager.scratchWorkspace", "true");
      }

    protected IndexedFileSystem indexedFileSystem;
    
    protected static void log (final @Nonnull String name)
      {
        logger.info("TEST: ****************************** " + name);  
      }
    
    @BeforeClass
    public static void setupLogging()
      {
        TestLoggerSetup.setupLogging(TestSupport.class);
      }
    
    @AfterClass
    public static void shutDownPersistence()
      {
        Locator.find(Persistence.class).shutdown();
      }

    @Before
    public void setupFiles()
      throws Exception 
      {
        eraseFiles();
        indexedFileSystem = new IndexedFileSystem();
        
        testFiles.clear();
        rootsTested = 0;
        
//        final File[] roots = File.listRoots(); // FIXME: on Windows it's important to test different roots... how?
        final File[] roots = new File[] { TEMP_FOLDER };
        
        for (final File root : roots)
          {
            if (!root.canWrite())
              {
                fail("Read only file system: " + root);
              }
            else
              {
                rootsTested++;
                final File baseFolder = new File(root, "BaseTestFolder");
                baseFolders.add(baseFolder);
                baseFolder.mkdirs();

                for (int j = 0; j < 100; j++)
                  {
                    final File file = new File(baseFolder, fileName(j));
                    file.getParentFile().mkdirs();
                    createContents(file, "Contents: " + j);
                    testFiles.add(file);
                  }
              }
          }
        
        assertFalse(testFiles.toString(), testFiles.isEmpty());
      }
        
    public void testMultiRootsOnWindows()
      {
        // On Windows you need two writable file systems. If you have only a single hard disk,
        // create a fake one with subst : e.g. mkdir C:\Fake, subst F: C:\Fake
        assertTrue("System is windows, but we're testing only 1 filesystem", Utilities.isWindows() == (rootsTested > 1));
      }
    
    protected void createContents (final @Nonnull File file, final @Nonnull String contents)
      throws IOException 
      {
        final PrintWriter pw = new PrintWriter(new FileWriter(file));
        pw.println(contents);
        pw.flush();
        pw.close();
      }
    
    protected static String fileName (final @Nonnegative int i)
      {
        return "testFolder" + (i / FOLDER_COUNT) + "/testFile" + i + "." + i; 
      }
    
    protected static String toPath (final @Nonnull FileObject fileObject)
      {
        assertNotNull(fileObject);
        
//        if (Utilities.isWindows())
          {
            return FileUtil.toFile(fileObject).getAbsolutePath();
          }

//        else // FIXME: check
//          {
//            return "/" + fileObject.getPath();  
//          }
      }
    
    protected static List<File> parents (final @Nonnull List<File> files)
      {
        final List<File> result = new ArrayList<File>();
        
        for (final File file : files)
          {
            final File parent = file.getParentFile();
            
            if (!result.contains(parent))
              {
                result.add(parent);  
              }
          }
        
        return result;
      }
    
    protected void eraseFiles() 
      {
        for (final File dir : baseFolders)
          {
            eraseDirectory(dir);
          }
      }
    
    private void eraseDirectory (final @Nonnull File dir)
      {
        for (final File child : dir.listFiles())
          {
            if (child.isDirectory())
              {
                eraseDirectory(child);  
              }
            
            child.delete();
          }
      }
  }
