/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: IndexedFileObjectTest.java,v 0959504d139a 2010/04/04 23:08:53 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.indexedfilesystem.impl;

import javax.annotation.Nonnull;
import java.util.Enumeration;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import org.openide.filesystems.FileAttributeEvent;
import org.openide.filesystems.FileChangeListener;
import org.openide.filesystems.FileEvent;
import org.openide.filesystems.FileLock;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileRenameEvent;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.indexedfilesystem.IndexedFileSystem;
import it.tidalwave.netbeans.indexedfilesystem.jpa.JPAFileIndexer;
import it.tidalwave.netbeans.indexedfilesystem.spi.FileIndexer;
import it.tidalwave.netbeans.indexedfilesystem.TestSupport;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;
import org.openide.filesystems.FileSystem;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: IndexedFileObjectTest.java,v 0959504d139a 2010/04/04 23:08:53 fabrizio $
 *
 **********************************************************************************************************************/
// @IssueTracker(type="jira", url="http://bluemarine.tidalwave.it/issues/browse")
public class IndexedFileObjectTest extends TestSupport 
  {
    static class MyListener implements FileChangeListener 
      {
        int fileFolderCreatedCount;
        int fileDataCreatedCount;
        int fileChangedCount;
        int fileDeletedCount;
        int fileRenamedCount;
        int fileAttributeChangedCount;
        Object latestEvent;
        
        @Override
        public void fileFolderCreated (final @Nonnull FileEvent fileEvent)
          {
            fileFolderCreatedCount++;
            latestEvent = fileEvent;
          }

        @Override
        public void fileDataCreated (final @Nonnull FileEvent fileEvent)
          {
            fileDataCreatedCount++;
            latestEvent = fileEvent;
          }

        @Override
        public void fileChanged (final @Nonnull FileEvent fileEvent)
          {
            fileChangedCount++;
            latestEvent = fileEvent;
          }

        @Override
        public void fileDeleted (final @Nonnull FileEvent fileEvent)
          {
            fileDeletedCount++;
            latestEvent = fileEvent;
          }

        @Override
        public void fileRenamed (final @Nonnull FileRenameEvent fileRenameEvent)
          {
            fileRenamedCount++;
            latestEvent = fileRenameEvent;
          }

        @Override
        public void fileAttributeChanged (final @Nonnull FileAttributeEvent fileAttributeEvent)
          {
            fileAttributeChangedCount++;
            latestEvent = fileAttributeEvent;
          }
        
        private void reset()
          {
            fileFolderCreatedCount = 0;
            fileDataCreatedCount = 0;
            fileChangedCount = 0;
            fileDeletedCount = 0;
            fileRenamedCount = 0;
            fileAttributeChangedCount = 0;
            latestEvent = null;
          }
      };
    
    private static final String TEST_ATTRIBUTE = "test.attribute";
    
    private static final MyListener fileChangeListener = new MyListener();
    
    private JPAFileIndexer fileIndexer = (JPAFileIndexer)Locator.find(FileIndexer.class);

    @Before
    public void setUpListener() 
      throws Exception 
      {
        indexedFileSystem.addFileChangeListener(fileChangeListener);
//        indexedFileSystem.getRoot().addFileChangeListener(fileChangeListener);
      }
    
/*    public void testAddFileChangeListener() 
      {
        log("addFileChangeListener");
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
      }*/

/*    public void testRemoveFileChangeListener() 
      {
        log("removeFileChangeListener");
        // TODO review the generated test code and remove the default call to fail.
        fail("The test case is a prototype.");
      }*/

    @Test
    public void mustBeAbleToReadID_ATTRIBUTE()
      {
        log("mustBeAbleToReadID_ATTRIBUTE");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            final Serializable id = (Serializable)fileObject.getAttribute(IndexedFileSystem.ID_ATTRIBUTE); // this initializes the id
            assertNotNull("File is not indexed: " + file, id);
          }
      }

    @Test
    public void mustBeAbleToSetAGenericAttribute()
      throws Exception 
      {
        log("mustBeAbleToSetAGenericAttribute");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            fileObject.setAttribute(TEST_ATTRIBUTE, file.getAbsolutePath());
          }
      }

    @Test
    public void mustBeAbleToGetAGenericAttribute()
      throws IOException
      {
        log("mustBeAbleToGetAGenericAttribute");
        setTestAttributes();

        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            assertEquals(file.getAbsolutePath(), fileObject.getAttribute(TEST_ATTRIBUTE));
          }
      }

    @Test
    public void mustBeAbleToEnumerateAttributes()
      throws IOException
      {
        log("mustBeAbleToEnumerateAttributes");
        setTestAttributes();

        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            final Enumeration e = fileObject.getAttributes();
            assertEquals(IndexedFileSystem.ID_ATTRIBUTE, e.nextElement());
            assertEquals(TEST_ATTRIBUTE, e.nextElement());
            assertFalse(e.hasMoreElements());
          }
      }

    // @IssueId("BM-281")
    @Test
    public void mustNotBeAbleToSetID_ATTRIBUTE()
      throws Exception 
      {
        log("mustNotBeAbleToSetID_ATTRIBUTE");
        
        for (File file : testFiles)
          {
            FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            try
              {
                fileObject.setAttribute(IndexedFileSystem.ID_ATTRIBUTE, "foo id");
              } 
            catch (IllegalArgumentException e)
              {
                continue; // ok, expected behaviour
              }
            
            fail("Could set ID_ATTRIBUTE");
          }
      }

/*    public void testEquals() 
      {
        log("equals");
        fail("The test case is a prototype.");
      }*/

/*    public void testHashCode() 
      {
        log("hashCode");
        fail("The test case is a prototype.");
      }*/

    @Test
    public void mustEnumerateChildrenAndChildrenMustBeIndexed()
      {
        log("mustEnumerateChildrenAndChildrenMustBeIndexed");
        
        for (final File file : parents(testFiles))
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            assertTrue(fileObject.isFolder());

            final FileObject[] children = fileObject.getChildren();
            assertNotNull(children);
            final File[] childrenFiles = file.listFiles();
            int delta = 0;
            
            if (childrenFiles[0].getName().equals(".nbattrs")) // TODO: explain why on Hudson .nbattrs is not there
              {
                delta = 1;  
              }
            
            assertEquals("" + file, childrenFiles.length - delta, children.length); // -delta --> ignores .nbattr
            
            for (int i = 0; i < childrenFiles.length - 1; i++)
              {
                assertEquals(childrenFiles[i + delta].getAbsolutePath(), toPath(children[i]));  
                assertNotNull("File is not indexed: " + children[i], children[i].getAttribute(IndexedFileSystem.ID_ATTRIBUTE));
              }
          }
      }

    @Test
    public void testGetChildrenIsBatched() 
      {
        log("getChildrenIsBatched");
        //
        // This also clears the internal cache of LocalFileSystems: in this way we will get
        // a fresh fileObject - otherwise we'd get a cached fileObject that's already populated
        // with children and no query would be performed.
        //
        fileIndexer.flushCaches();
            
        for (File file : parents(testFiles))
          {
            FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            assertTrue(fileObject.isFolder());
            // Scanning parent directories of fileObject has probably filled most of the pendingList:
            // flush it to be sure that getChildren() will start with a clean status.
            fileIndexer.flushPending();

            int queryCount = fileIndexer.queryCount;
            FileObject[] children = fileObject.getChildren();
            assertNotNull(children);
            File[] childrenFiles = file.listFiles();
            int delta = 0;
            
            if (childrenFiles[0].getName().equals(".nbattrs")) // TODO: explain why on Hudson .nbattrs is not there
              {
                delta = 1;  
              }
            
            assertEquals("" + file, childrenFiles.length - delta, children.length); // -delta --> ignores .nbattr
            
            for (int i = 0; i < childrenFiles.length - 1; i++)
              {
                assertEquals(childrenFiles[i + delta].getAbsolutePath(), toPath(children[i]));  
                assertNotNull("File is not indexed: " + children[i], children[i].getAttribute(IndexedFileSystem.ID_ATTRIBUTE));
              }

            int deltaQuery = fileIndexer.queryCount - queryCount;
            assertEquals("Wasn't executed in batch", 1, deltaQuery);
          }
      }

    @Test
    public void testCreateFolder() 
      throws Exception 
      {
        log("createFolder");
        int i = 0;
        
        for (File file : testFiles)
          {
            FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            
            if (fileObject.isFolder())
              {
                FileObject folder = fileObject.getParent().createFolder("new folder " + i++);
                assertNotNull("File is not indexed", folder.getAttribute(IndexedFileSystem.ID_ATTRIBUTE));
                assertEquals(fileObject, folder.getParent());
              }
            //
            // Now try to create a subfolder of a file - should fail
            //
            try 
              {
                FileObject folder = fileObject.createFolder("new folder");
              } 
            catch (IOException e) 
              {
                continue;
              }

            fail("Exception expected");
          }
      }

/*    public void testCreateData() 
      throws Exception 
      {
        log("createData");
        
        for (File file : files)
          {
            FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
          }
        
        fail("The test case is a prototype.");
      }*/

    public void mustBeAbleToNavigateToParentAndParentMustBeIndexed()
      {
        log("mustBeAbleToNavigateToParentAndParentMustBeIndexed");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            final FileObject parent = fileObject.getParent();
            assertNotNull("File is not indexed: " + parent, parent.getAttribute(IndexedFileSystem.ID_ATTRIBUTE));
            assertEquals(file.getParentFile().getAbsolutePath(), toPath(parent));
          }
      }

    @Test // Note returns the inner LocalIndexedFileSystem instead of IndexedFileSystem
    public void testGetFileSystem() 
      throws Exception 
      {
        log("getFileSystem");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            final FileSystem fileSystem = fileObject.getFileSystem();
            assertTrue(fileSystem instanceof LocalIndexedFileSystem);
          }
      }

    @Test
    public void mustBeNamedAsTheRealFile()
      {
        log("mustBeNamedAsTheRealFile");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            String name = file.getName();
            final int i = name.lastIndexOf('.');
            assertTrue(i > 0);
            name = name.substring(0, i);
            assertEquals(name, fileObject.getName());
          }
      }

    @Test
    public void mustHaveTheSameExtensionAsTheRealFile()
      {
        log("mustHaveTheSameExtensionAsTheRealFile");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            String name = file.getName();
            final int i = name.lastIndexOf('.');
            assertTrue(i > 0);
            final String ext = name.substring(i + 1);
            assertEquals(ext, fileObject.getExt());
          }
      }

    @Test
    public void mustBeAbleToDistinguishFoldersFromPlainFiles()
      {
        log("mustBeAbleToDistinguishFoldersFromPlainFiles");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            assertEquals(file.isDirectory(), fileObject.isFolder());
          }
      }

//    @Test  FIXME: fails on Hudson, with a delta = 4000
    public void testLastModified() 
      {
        log("lastModified");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
//            assertEquals(file.lastModified(), fileObject.lastModified().getTime()); // FIXME: why?
            long delta = Math.abs(file.lastModified() - fileObject.lastModified().getTime());
            assertTrue("Delta for " + file + ": " + delta, delta <= 1000L);
          }
      }

    @Test
    public void mustBeAbleToDistinguishRoots()
      {
        log("mustBeAbleToDistinguishRoots");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            assertFalse(fileObject.isRoot());
          }
        // TODO: add a test for the real root
      }

    @Test
    public void testIsData() 
      {
        log("isData");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            assertTrue(fileObject.isData());
          }
      }

    @Test
    public void testIsValid() 
      {
        log("isValid");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            assertTrue(fileObject.isValid());
          }
      }

    @Test
    public void mustReturnTheSizeOfTheRealFile()
      {
        log("mustReturnTheSizeOfTheRealFile");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            assertEquals(file.length(), fileObject.getSize());
          }
      }

    @Test
    public void mustBeAbleToWriteData()
      throws Exception 
      {
        log("mustBeAbleToWriteData");
        
        int i = 0;
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            final FileLock lock = fileObject.lock();
            final OutputStream os = fileObject.getOutputStream(lock);
            assertNotNull(os);
            final DataOutputStream dos = new DataOutputStream(os);
            dos.writeInt(0x01234567);
            dos.writeInt(i++);
            dos.close();
            lock.releaseLock();
          }

        i = 0;
        for (final File file : testFiles)
          {
            final DataInputStream dis = new DataInputStream(new FileInputStream(file));
            assertEquals("Unexpected value reading " + file, 0x01234567, dis.readInt());
            assertEquals("Unexpected value reading " + file, i++, dis.readInt());
            dis.close();
          }
      }

    @Test
    public void mustBeAbleToReadData()
      throws Exception 
      {
        log("mustBeAbleToReadData");
        
        int i = 0;
        for (final File file : testFiles)
          {
            final DataOutputStream dos = new DataOutputStream(new FileOutputStream(file));
            dos.writeInt(0x01234567);
            dos.writeInt(i++);
            dos.close();
          }

        i = 0;
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            final InputStream is = fileObject.getInputStream();
            assertNotNull(is);
            final DataInputStream dis = new DataInputStream(is);
            assertEquals("Unexpected value reading " + fileObject, 0x01234567, dis.readInt());
            assertEquals("Unexpected value reading " + fileObject, i++, dis.readInt());
            dis.close();
          }
      }

    @Test
    public void testLock() 
      throws Exception 
      {
        log("lock");
        
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            final FileLock lock = fileObject.lock();
            // FIXME: something to assert?
            lock.releaseLock();
          }
      }
    
    @Test
    public void mustBeAbleToDeleteIndexedFiles()
      throws Exception 
      {
        log("mustBeAbleToDeleteIndexedFiles");
        // With the first directory we test recursive deletion
        final File directoryForRecursiveDeletion = testFiles.get(0).getParentFile();
        FileObject fileObject = indexedFileSystem.findResource(directoryForRecursiveDeletion.getAbsolutePath());
        fileObject.delete();
        assertFalse("Parent directory has been removed", directoryForRecursiveDeletion.exists());
        assertFalse("Indexing is still here", fileIndexer.isIndexed(directoryForRecursiveDeletion.getAbsolutePath()));
        
        for (final File file : testFiles)
          {
            if (file.getParentFile().equals(directoryForRecursiveDeletion))
              {
                assertFalse("File has not been deleted: " + file, file.exists());
                assertFalse("Indexing is still here for: " + file, fileIndexer.isIndexed(file.getAbsolutePath()));
              }
          }
        
        for (final File file : testFiles)
          {
            if (!file.getParentFile().equals(directoryForRecursiveDeletion))
              {
                fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
                log("delete: " + file + " = " + fileObject + " = " + file.exists());
                fileObject.delete();
                log("delete: " + file + " = " + file.exists());
                assertFalse("File has not been deleted: " + file, file.exists());
                assertFalse("Indexing is still here for: " + file, fileIndexer.isIndexed(file.getAbsolutePath()));
              }
          }
      }
    
    @Test
    public void mustBeAbleToDeleteFilesNotIndexedYet()
      throws Exception
      {
        log("mustBeAbleToDeleteFilesNotIndexedYet");
        final File testFile = new File(TEMP_FOLDER, "test_file");
        createContents(testFile, "Contents");
        assertTrue("File cannot be created", testFile.exists());
        final FileObject fileObject = indexedFileSystem.findResource(testFile.getAbsolutePath());
        // at this point the id has not been created yet as it hasn't been requested yet.
        assertNotNull("Null fileObject for " + testFile, fileObject);
        fileObject.delete();
        assertFalse("Indexing is still here", fileIndexer.isIndexed(testFile.getAbsolutePath()));
        assertFalse("File has not been deleted", testFile.exists());
      }
    
    @Test
    public void testRenameFile() 
      throws Exception 
      {
        log("renameFile");
        fileIndexer.flushCaches();
              
        for (final File file : testFiles)
          {
            if (file.isFile())
              {
                log("Rename " + file);
                assertTrue("File doesn't exist: " + file, file.exists());
                String path = file.getAbsolutePath();
                final FileObject fileObject = indexedFileSystem.findResource(path);
                assertNotNull("Null fileObject for " + path, fileObject);
                final Serializable id = (Serializable)fileObject.getAttribute(IndexedFileSystem.ID_ATTRIBUTE);
                assertNotNull("File not indexed: " + fileObject, id);
                final String newName = fileObject.getName() + "-NEW";
                final String ext = fileObject.getExt();
                
                fileChangeListener.reset();
                final FileLock lock = fileObject.lock();
                fileObject.rename(lock, newName, ext);
                lock.releaseLock();
                
                final String newNameWithExt = newName + "." + ext;
                assertEquals("The final name is not updated", newNameWithExt, fileObject.getNameExt());
//                assertEquals("Listener not invoked", 1, fileChangeListener.fileRenamedCount);
                
                path = fileIndexer.findPathById(id);
                assertNotNull("Cannot find path by id ", path);
                assertEquals("IndexedFileSystem has still the old path", toPath(fileObject), path);
                assertFalse("The new file is not indexed", fileIndexer.isIndexed(file.getAbsolutePath()));
              }
          }
      }

    @Test
    public void testRenameFolder() 
      throws Exception 
      {
        log("renameFolder");
        fileIndexer.flushCaches();
              
        for (final File file : parents(testFiles))
          {
            assertTrue(file.isDirectory());
            log("Rename " + file);
            assertTrue("File doesn't exist: " + file, file.exists());
            String path = file.getAbsolutePath();
            FileObject fileObject = indexedFileSystem.findResource(path);
            assertNotNull("Null fileObject for " + path, fileObject);
            Serializable oldId = (Serializable)fileObject.getAttribute(IndexedFileSystem.ID_ATTRIBUTE);
            assertNotNull("File not indexed: " + fileObject, oldId);
            String newName = fileObject.getName() + "-NEW";
            String ext = fileObject.getExt();

            fileChangeListener.reset();
            FileLock lock = fileObject.lock();
            log(String.format(">>>> newName: %s, ext: %s ", newName, ext));
            fileObject.rename(lock, newName, ext);
            lock.releaseLock();

            String newNameWithExt = newName + (ext.equals("") ? "" : ("." + ext));
            assertEquals("The final name is not updated: " + fileObject, newNameWithExt, fileObject.getNameExt());
//                assertEquals("Listener not invoked", 1, fileChangeListener.fileRenamedCount);

            path = fileIndexer.findPathById(oldId);
            assertNotNull("Cannot find path by id ", path);
            assertEquals("IndexedFileSystem has still the old path", toPath(fileObject), path);
            assertFalse("The new file is not indexed", fileIndexer.isIndexed(file.getAbsolutePath()));
            
            // TODO check all children
          }
      }

    private void setTestAttributes()
      throws IOException
      {
        for (final File file : testFiles)
          {
            final FileObject fileObject = indexedFileSystem.findResource(file.getAbsolutePath());
            fileObject.setAttribute(TEST_ATTRIBUTE, file.getAbsolutePath());
          }
      }
  }
