/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: IndexedFileSystemTest.java,v 156ca56ad125 2010/04/04 22:34:31 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.indexedfilesystem.impl;

import it.tidalwave.netbeans.indexedfilesystem.jpa.JPAFileIndexer;
import it.tidalwave.netbeans.indexedfilesystem.spi.FileIndexer;
import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.io.IOException;
import java.io.Serializable;
import org.openide.util.Utilities;
import org.openide.util.actions.SystemAction;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileStateInvalidException;
import org.openide.filesystems.FileUtil;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.indexedfilesystem.TestSupport;
import it.tidalwave.netbeans.indexedfilesystem.IndexedFileSystem;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: IndexedFileSystemTest.java,v 156ca56ad125 2010/04/04 22:34:31 fabrizio $
 *
 **********************************************************************************************************************/
public class IndexedFileSystemTest extends TestSupport
  {
    private static final Map<String, Serializable> idMapByPath = new HashMap<String, Serializable>();
    
    private static final Map<Serializable, String> pathMapById = new HashMap<Serializable, String>();
    
    private JPAFileIndexer fileIndexer = (JPAFileIndexer)Locator.find(FileIndexer.class);
    
    @Before
    public void clearStuff()
      {
        idMapByPath.clear();
        pathMapById.clear();
      }
    
    @Test
    public void testGetRoot() 
      {
        log("getRoot");
        FileObject result = indexedFileSystem.getRoot();
        assertEquals("", result.getPath()); // "", not "/" since path is relative to the root
      }

    @Test
    public void testGetDisplayName() 
      {
        log("getDisplayName");
        assertEquals("My Computer", indexedFileSystem.getDisplayName());
      }

    @Test
    public void testIsReadOnly() 
      {
        log("isReadOnly");
        assertFalse(indexedFileSystem.isReadOnly());
      }

    @Test
    public void testGetActions() 
      {
        log("getActions");
        SystemAction[] result = indexedFileSystem.getActions();
        assertEquals(0, result.length);
//        assertEquals(1, result.length);
      }  

    @Test
    public void testFindResource()
      throws IOException
      {
        log("findResource");
        Set<Object> idSet = new HashSet<Object>();
        
        for (File file : testFiles)
          {
            String path = file.getAbsolutePath();
//            
//            if (path.startsWith("F:"))
//            {
//                int x = 3;
//            }
            FileObject fileObject = indexedFileSystem.findResource(path);
            assertEquals(file.getAbsolutePath(), toPath(fileObject));
            Serializable id = (Serializable)fileObject.getAttribute(IndexedFileSystem.ID_ATTRIBUTE);
            assertNotNull("File is not indexed: " + file, id);
            assertFalse("Duplicated id: " + id + " " + file, idSet.contains(id));
            idSet.add(id);
            idMapByPath.put(path, id);
            pathMapById.put(id, path);
          }
      }
    
    @Test
    public void testFindResourceIsPersistent()
      throws IOException
      {
        log("findResourceIsPersistent");
        testFindResource();
        
        for (File file : testFiles)
          {
            String path = file.getAbsolutePath();
            FileObject fileObject = indexedFileSystem.findResource(path);
            assertEquals(file.getAbsolutePath(), toPath(fileObject));
            Serializable id = (Serializable)fileObject.getAttribute(IndexedFileSystem.ID_ATTRIBUTE);
            assertNotNull("File is not indexed: " + file, id);
            assertEquals("Id was not persistent: " + id, idMapByPath.get(file.getAbsolutePath()), id);
          }
      }
    
//    @Test FIXME: doesn't work'
    public void testToFileObject() 
      throws IOException 
      {
        log("testToFileObject");
        testFindResource();
        
        for (File file : testFiles)
          {
            FileObject fileObject = FileUtil.toFileObject(file);
            assertEquals(file.getAbsolutePath(), toPath(fileObject)); 
            Serializable id = (Serializable)fileObject.getAttribute(IndexedFileSystem.ID_ATTRIBUTE);
            assertNotNull("File is not indexed: " + file, id);
            assertEquals("Id was not persistent: " + id, idMapByPath.get(file.getAbsolutePath()), id);
          }
      } 

    @Test
    public void testIsIndexed() 
      {
        log("isIndexed");
        
        for (File file : testFiles)
          {
            String path = file.getAbsolutePath();
            assertTrue("File is not indexed: " + file, fileIndexer.isIndexed(path));
          }
      }

    @Test
    public void testFindAllIndexedIds() 
      throws IOException 
      {
        log("findAllIndexedIds");
        testFindResource();
        
        Collection<Serializable> ids = indexedFileSystem.findAllIndexedIds();
        assertEquals(testFiles.size(), ids.size());
        
        for (Serializable id : pathMapById.keySet())
          {
            assertTrue(ids.contains(id));  
          }
      }

    @Test
    public void testFindRoots() 
      throws FileStateInvalidException
      {
        log("testFindRoots");
        
        File[] roots = File.listRoots();
        Collection<FileObject> rootFileObjects = indexedFileSystem.findRoots();
        assertEquals(roots.length, rootFileObjects.size());

        Iterator<FileObject> i = rootFileObjects.iterator();
        
        for (File root : roots)
          {
            FileObject fileObject = i.next();
            Serializable id = (Serializable)fileObject.getAttribute(IndexedFileSystem.ID_ATTRIBUTE);
            assertNotNull("File is not indexed: " + root, id);
            assertEquals(root.getAbsolutePath(), toPath(fileObject));
          }
      }

    @Test
    public void testFindResourceById() 
      throws IOException 
      {
        log("findResourceById");
        testFindResource();

        for (Serializable id : pathMapById.keySet())
          {
            String path = pathMapById.get(id);
            FileObject fileObject = indexedFileSystem.findResourceById(id);
            assertEquals(path, toPath(fileObject));
            Serializable id2 = (Serializable)fileObject.getAttribute(IndexedFileSystem.ID_ATTRIBUTE);
            assertNotNull("Missing id" + id2);
            assertEquals("Id mismatch", id, id2);
          }
      }

    @Test
    public void testFindPathById() 
      throws IOException 
      {
        log("findPathById");
        testFindResource();
        
        for (Serializable id : pathMapById.keySet())
          {
            String path = pathMapById.get(id);
            assertEquals(path, fileIndexer.findPathById(id));
          }
      }

    @Test
    public void testFindIdByPath() 
      throws IOException 
      {
        log("findIdByPath"); 
        testFindResource();
        
        for (String path : idMapByPath.keySet())
          {
            Serializable id = idMapByPath.get(path);
            
            if (path.startsWith("/"))
              {
                path = path.substring(1);
              }
            
            assertEquals(id, fileIndexer.findIdByPath(path));
          }
      }

    @Test
    public void testRemoveIndexing() 
      {
        log("removeIndexing");
        
        for (File file : testFiles)
          {
            FileObject fileObject = FileUtil.toFileObject(file);
            String path = file.getAbsolutePath();
            indexedFileSystem.removeIndexing(fileObject);
            assertFalse(fileIndexer.isIndexed(path));
          }
      }

    @Test
    public void testNormalizedPathOnUnix() 
      {
        log("testNormalizedPathOnUnix");
        
        if (!Utilities.isWindows())
          {
            assertEquals(fileIndexer.normalizedPath("/"), "/");
            assertEquals(fileIndexer.normalizedPath("/Foo/Bar"), "/Foo/Bar");
            assertEquals(fileIndexer.normalizedPath("/Foo/Bar/"), "/Foo/Bar");
          }
      }

    @Test
    public void testLocalizedPathOnUnix() 
      {
        log("testLocalizedPathOnUnix");
        
        if (!Utilities.isWindows())
          {
            assertEquals(fileIndexer.localizedPath("/"), "/");
            assertEquals(fileIndexer.localizedPath("/Foo/Bar"), "/Foo/Bar");
          }
      }    
    
    @Test
    public void testNormalizedPathOnWindows() 
      {
        log("testNormalizedPathOnWindows");
        
        if (Utilities.isWindows())
          {
            assertTrue("No secondary drive, can't do a complete test", File.listRoots().length >= 2);
            
            for (File root : File.listRoots())
              {
                String drive = root.getPath().substring(0, 1);
                
                if (drive.equals("C"))
                  {
                    assertEquals("For " + root, "/", fileIndexer.normalizedPath("C:"));
                    assertEquals("For " + root, "/", fileIndexer.normalizedPath("C:\\"));
                    assertEquals("For " + root, "/Foo/Bar", fileIndexer.normalizedPath("C:\\Foo\\Bar"));
                    assertEquals("For " + root, "/Foo/Bar", fileIndexer.normalizedPath("C:\\Foo\\Bar\\"));
                  }
                else
                  {
                    assertEquals("For " + root, drive + ":", fileIndexer.normalizedPath(drive + ":"));
                    assertEquals("For " + root, drive + ":", fileIndexer.normalizedPath(drive + ":\\"));
                    assertEquals("For " + root, drive + ":/Foo/Bar", fileIndexer.normalizedPath(drive + ":\\Foo\\Bar"));
                    assertEquals("For " + root, drive + ":/Foo/Bar", fileIndexer.normalizedPath(drive + ":\\Foo\\Bar\\"));
                  }
              }
          }
      }

    @Test
    public void testLocalizedPathOnWindows() 
      {
        log("testLocalizedPathOnWindows");
        
        if (Utilities.isWindows())
          {
            assertTrue("No secondary drive, can't do a complete test", File.listRoots().length >= 2);
            
            for (File root : File.listRoots())
              {
                String drive = root.getPath().substring(0, 1);
                
                if (drive.equals("C"))
                  {
                    assertEquals(fileIndexer.localizedPath("/"), "C:\\");
                    assertEquals(fileIndexer.localizedPath("/Foo/Bar"), "C:\\Foo\\Bar");
                  }
                else
                  {
                    assertEquals(fileIndexer.localizedPath(drive + ":"), drive + ":");
                    assertEquals(fileIndexer.localizedPath(drive + ":/Foo/Bar"), drive + ":\\Foo\\Bar");
                  }
              }
          }
      }    
  }
