/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: RendezVous.java,v 4e8a394a80a9 2010/03/28 20:51:11 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.rendezvous;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.awt.EventQueue;
import it.tidalwave.util.logging.Logger;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: RendezVous.java,v 4e8a394a80a9 2010/03/28 20:51:11 fabrizio $
 *
 **********************************************************************************************************************/
public class RendezVous
  {
    private static final String CLASS = RendezVous.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    private final Lock lock = new ReentrantLock();

    private final Map<String, Condition> conditionMap = new HashMap<String, Condition>();

    private final Map<String, AtomicInteger> countMap = new HashMap<String, AtomicInteger>();

    private final Map<String, Set<Tag>> tagMap = new HashMap<String, Set<Tag>>();

    private final List<RendezVous> children = new ArrayList<RendezVous>();

    public final static class Tag
      {
        private static int idGenerator;
        
        private final int id = idGenerator++;

        @Override
        public boolean equals (@CheckForNull final Object object)
          {
            if (object == null)
              {
                return false;
              }

            if (getClass() != object.getClass())
              {
                return false;
              }

            final Tag other = (Tag)object;
            return this.id == other.id;
          }

        @Override
        public int hashCode()
          {
            return id;
          }

        @Override
        public String toString()
          {
            return String.format("Tag[%d]", id);
          }

        private final static ThreadLocal<Tag> THREAD_LOCAL = new ThreadLocal<Tag>();

        @CheckForNull
        public static Tag get()
          {
            return THREAD_LOCAL.get();
          }

        public static void set (@Nullable final Tag tag) // FIXME: make protected
          {
            logger.fine("Tagging %s with %s...", Thread.currentThread(), tag);
            THREAD_LOCAL.set(tag);
          }
      }

    @Nonnull
    public static Tag invokeLaterTagged (@Nonnull final Runnable runnable)
      {
        final Tag tag = new Tag();
        invokeLater(runnable, tag);
        return tag;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static void invokeLater (@Nonnull final Runnable runnable)
      {
        invokeLater(runnable, Tag.get());
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    private static void invokeLater (@Nonnull final Runnable runnable,
                                     @Nonnull final Tag tag)
      {
        EventQueue.invokeLater(new Runnable()
          {
            public void run()
              {
//                try
//                  {
                    Tag.set(tag);
                    runnable.run();
//                  }
//                finally
//                  {
//                    Tag.set(null);
//                  }
              }
          });

        // It's better to post a new event to remove the tag, since the
        // runnable could have posted events on the queue on its own, and
        // we want that they are executed while the EDT is still tagged.
        EventQueue.invokeLater(new Runnable()
          {
            public void run()
              {
                Tag.set(null);
              }
          });
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public RendezVous (@Nonnull final String ... eventNames)
      {
        for (final String eventName : eventNames)
          {
            findOrCreateCondition(eventName);
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public RendezVous bind (@Nonnull final RendezVous ... rendezVouses)
      {
        children.addAll(Arrays.asList(rendezVouses));
        return this;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public Waitable wait (@Nonnull final String eventName)
      {
        Condition condition = conditionMap.get(eventName);

        if (condition == null)
          {
            for (final RendezVous child : children)
              {
                condition = child.conditionMap.get(eventName);

                if (condition != null)
                  {
                    return new Waitable(child, condition, eventName, child.countMap.get(eventName).intValue());
                  }
              }

            throw new IllegalArgumentException(String.format("Invalid condition name: %s - %s", eventName, conditionMap.keySet()));
          }

        return new Waitable(this, condition, eventName, countMap.get(eventName).intValue());
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    protected void waitForEvent (@Nonnull final String eventName,
                                 @Nonnull final Condition condition,
                                 @Nonnegative final long timeout,
                                 @CheckForNull final Tag tag)
      throws InterruptedException
      {
        logger.fine("waitForEvent(%s, %d, %s)", eventName, timeout, tag);

        lock.lock();

        try
          {
            while (!matchesTag(eventName, tag))
              {
                if (!condition.await(timeout, TimeUnit.MILLISECONDS))
                  {
                    logger.finer(">>>> %s: timeout expired", eventName);
                    throw new InterruptedException("Timeout expired: " + eventName);
                  }
              }

            logger.finer(">>>> %s triggered", eventName);
          }
        finally
          {
            lock.unlock();
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    private boolean matchesTag (@Nonnull final String eventName,
                                @CheckForNull final Tag tag)
      {
        final Set<Tag> tags = tagMap.get(eventName);
        logger.finest(">>>> current event tagged as %s", tags);
        return (tag == null) ? true : tags.contains(tag);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public void notifyEventOccurred (@Nonnull final String eventName)
      {
        logger.fine("notifyEventOccurred(%s) - tag: %s", eventName, Tag.get());

        lock.lock();

        try
          {
            final int count = countMap.get(eventName).incrementAndGet();
            final Tag tag = Tag.get();

            if (tag != null)
              {
                tagMap.get(eventName).add(tag);
              }

            logger.fine(">>>> %s count is %d", eventName, count);
            findOrCreateCondition(eventName).signalAll();
          }
        finally
          {
            lock.unlock();
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    private Condition findOrCreateCondition (@Nonnull final String eventName)
      {
        Condition condition = conditionMap.get(eventName);

        if (condition == null)
          {
            condition = lock.newCondition();
            conditionMap.put(eventName, condition);
            countMap.put(eventName, new AtomicInteger(0));
            tagMap.put(eventName, Collections.synchronizedSet(new HashSet<Tag>()));
          }

        return condition;
      }
  }
