/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: PersistenceImpl.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.persistence.impl;

import javax.annotation.Nonnull;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Properties;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import javax.sql.DataSource;
import java.io.Serializable;
import javax.persistence.Column;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceException;
import javax.persistence.Table;
import com.eaio.uuid.UUID;
import org.hibernate.ejb.HibernateEntityManager;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.persistence.Persistence;
import it.tidalwave.netbeans.persistence.TxTask;
import it.tidalwave.netbeans.persistence.spi.PersistenceConfigurator;
import org.openide.util.lookup.ServiceProvider;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: PersistenceImpl.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
@ServiceProvider(service=Persistence.class)
public class PersistenceImpl implements Persistence
  {
    private static final String CLASS = PersistenceImpl.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    
    private final Properties properties;
    
    private EntityManagerFactory entityManagerFactory;
    
//    public static final PrintStream databaseStream = new PrintStream(new LoggerPrintStream());
    
    private boolean metadataRetrieved = false;
    
    private final List<String> tableNames = new ArrayList<String>();

    private final PersistenceConfigurator configurator = Locator.find(PersistenceConfigurator.class);

    private final DataSource dataSource = new DataSourceImpl();
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public PersistenceImpl()
      {
        properties = configurator.initialize();
        
//        try
//          {
//            final InitialContext context = new InitialContext();
//            context.bind(Persistence.DATASOURCE_NAME, new DataSourceImpl());
//          }
//        catch (NamingException e)
//          {
//            logger.throwing(CLASS, "ctor()", e);
//            logger.warning("Couldn't bind the DataSource");
//          }
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override @Nonnull
    public DataSource getDataSource()
      {
        return dataSource;
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public void shutdown()
      {
        configurator.shutdown();
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public Properties getProperties()
      {
        return (Properties)properties.clone();
      }
    
    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override @Nonnull
    public synchronized EntityManagerFactory getEntityManagerFactory()
      {
        if (entityManagerFactory == null)
          {
            entityManagerFactory = javax.persistence.Persistence.createEntityManagerFactory(configurator.getPersistenceUnitName(), properties);
          }
        
        return entityManagerFactory;
      }
    
    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public synchronized void reset()
      {
        logger.info("reset()");
        
        if (entityManagerFactory != null)
          {
            entityManagerFactory.close();
            entityManagerFactory = null;
          }
        
        getEntityManagerFactory();
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override
    public void createIndexes (final Class<?> entityClass, final String ... fieldNames)
      throws IllegalArgumentException
      {
        logger.info("createIndexes(%s, %s)", entityClass, Arrays.toString(fieldNames));
        
        final Table table = entityClass.getAnnotation(Table.class);
        final String tableName = (table != null) ? table.name() : entityClass.getName().replaceAll(".*\\.", "");
        
        try
          {
            TxTask.execute(NoSuchFieldException.class, new TxTask<Void, NoSuchFieldException>() 
              {
                @Override
                public Void run() 
                  throws NoSuchFieldException
                  {
                    final EntityManager entityManager = getEntityManager();
                    // Reference the entity, so you're sure that the related table has been created
                    entityManager.find(entityClass, "");
                    
loop:               for (final String fieldName : fieldNames)
                      {
                        for (final Field field : entityClass.getDeclaredFields())
                          {
                            if (fieldName.equals(field.getName()))
                              {
                                final Column column = field.getAnnotation(Column.class);
                                final String columnName = (column != null) ? column.name() : field.getName();
                                final String ddl = String.format("CREATE INDEX %s_INDEX ON %s (%s)", columnName, tableName, columnName);
                                logger.info(">>>> creating index: %s", ddl);
                                final long time = System.currentTimeMillis();
                                
                                if (!hasIndex(tableName, columnName))
                                  {
                                    entityManager.createNativeQuery(ddl).executeUpdate();
                                    logger.info(">>>>>>>> index created in %d msec", System.currentTimeMillis() - time);
                                  }
                                else
                                  {
                                    logger.info(">>>> index already exists");
                                  }

                                continue loop;
                              }
                          }

                        throw new NoSuchFieldException(fieldName);
                      }

                    return null;
                  }
              });
          }
        catch (NoSuchFieldException e)
          {
            throw new IllegalArgumentException(e);
          }
      }
    
    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    public boolean tableExists (final String tableName)
      {
        retrieveDatabaseMetadata();
        return tableNames.contains(tableName);
      }
    
    /***************************************************************************
     *
     * {@inheritDoc}
     *
     **************************************************************************/
    @Override @Nonnull
    public Serializable createUniqueId()
      {
        return new UUID().toString();
      }
    
    /***************************************************************************
     *
     * Retrieves some metadata of the database.
     *
     **************************************************************************/
    public DatabaseMetaData getDatabaseMetaData() 
      throws SQLException
      {
        final EntityManager em = TxTask.getEntityManager();
        final Connection conn = getJdbcConnection(em);
        return conn.getMetaData();
      }
    
    /***************************************************************************
     *
     * Finds out if a given column in a given table has an index.
     * 
     * @param   tableName    the table name
     * @param   columnName   the column name
     * @return  true if the index already exists
     *
     **************************************************************************/
    private boolean hasIndex (final String tableName, final String columnName) 
      {
        boolean result = false;
        
        try
          {
            final DatabaseMetaData databaseMetaData = getDatabaseMetaData();
            final ResultSet rs = databaseMetaData.getIndexInfo(null, null, tableName, false, true);

            while (rs.next())
              {
                logger.finest(">>>> %s $s", rs.getString("INDEX_NAME"), rs.getString("COLUMN_NAME"));
                
                if (columnName.equals(rs.getString("COLUMN_NAME")))
                  {
                    result = true;
                    break;
                  }
              }
            
            rs.close();
          }
        catch (SQLException e)
          {
          }
        
        logger.fine(">>>> %s.%s has index: %s", tableName, columnName, result);
      
        return result;
      }
                
    /***************************************************************************
     *
     * Retrieves some metadata of the database.
     *
     **************************************************************************/
    private synchronized void retrieveDatabaseMetadata()
      {
        if (!metadataRetrieved)
          {
            try
              {
                TxTask.execute(SQLException.class, new TxTask<Void, SQLException>() 
                  {
                    @Override
                    public Void run()
                      throws SQLException
                      {
                        final DatabaseMetaData databaseMetaData = getDatabaseMetaData();
                        final ResultSet resultSet = databaseMetaData.getTables("APP", null, null, null);

                        while (resultSet.next())
                          {
                            tableNames.add(resultSet.getString("TABLE_NAME"));
                          }

                        resultSet.close();
                        logger.info("Database tables: %s", tableNames);
                        return null;
                      }
                  });
              }
            catch (SQLException e)
              {
                throw new PersistenceException(e);
              }
              
            metadataRetrieved = true;
          }
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    protected Connection getJdbcConnection (final EntityManager em)
      {
        return ((HibernateEntityManager)em).getSession().connection();  
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
//    private String upgradeTo_0_9_1a (String databaseVersion, final Class clazz) 
//      throws HibernateException, SQLException 
//      {
//        if (clazz.getName().endsWith("ThumbnailPB") && databaseVersion.equals(VERSION_0_9_1))
//          {
//            // Changes suppressed since they are obsolete
//            setDatabaseVersion(databaseVersion = VERSION_0_9_1a);
//          }
//        
//        return databaseVersion;
//      }
  }
