/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: GenericCapabilityActionTest.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.util.actions;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import javax.annotation.Nonnull;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.Test;
import org.mockito.InOrder;
import static org.mockito.Mockito.*;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.util.actions.CookieAction;
import org.openide.util.lookup.Lookups;

class Stuff
  {
  }

class StuffSubClass extends Stuff
  {
  }


class GenericCapabilityActionFixture<T> extends GenericCapabilityAction<T>
  {
    public GenericCapabilityActionFixture (final @Nonnull Class<? extends T> ... capabilityClasses)
      {
        super(capabilityClasses);
      }

    @Override
    protected void performAction (final @Nonnull List<T> capabilities)
      {
      }
  }

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: GenericCapabilityActionTest.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
public class GenericCapabilityActionTest
  {
    private final Object object1 = new Object();
    private final Object object2 = new Object();
    private final Stuff stuff1 = new Stuff();
    private final Stuff stuff2 = new StuffSubClass();
    private final List<Object> capabilities = Arrays.asList(object1, stuff1, object2, stuff2);
    private final Node[] nodes = createNodes(capabilities);

    private GenericCapabilityAction<Object> objectAction;
    private GenericCapabilityAction<Stuff> stuffAction;
    private GenericCapabilityAction<StuffSubClass> subStuffAction;

    @Before
    public void setUp() 
      {
        objectAction = new GenericCapabilityActionFixture<Object>(Object.class);
        stuffAction = new GenericCapabilityActionFixture<Stuff>(Stuff.class);
        subStuffAction = new GenericCapabilityActionFixture<StuffSubClass>(StuffSubClass.class);
      }

    @Test
    public void shouldReturnOnlySpecificCapabilityInstances()
      {
        assertArrayEquals(new Object[]{object1, stuff1, object2, stuff2}, objectAction.getCapabilities(nodes).toArray());
        assertArrayEquals(new Object[]{         stuff1,          stuff2}, stuffAction.getCapabilities(nodes).toArray());
        assertArrayEquals(new Object[]{                          stuff2}, subStuffAction.getCapabilities(nodes).toArray());
      }

    @Test
    public void shouldInvokePerformActionOnAllCapabilityInstancesExtractedFromNodes()
      throws Exception
      {
        final GenericCapabilityAction<Object> fixture = spy(objectAction);
        fixture.performAction(createNodes(capabilities));
        verify(fixture).performAction(eq(capabilities));
        verifyNoMoreInteractions(fixture);
      }


    @Test
    public void shouldInvokeEnableOnAllCapabilityInstancesExtractedFromNodes()
      throws Exception
      {
        final GenericCapabilityAction<Object> fixture = spy(objectAction);
        final InOrder inOrder = inOrder(fixture);
        
        assertTrue(fixture.enable(nodes));

        inOrder.verify(fixture).enable(same(nodes));

        for (final Object capability : capabilities)
          {
            inOrder.verify(fixture).enable(same(capability));
          }

//        verifyNoMoreInteractions(fixture); // FIXME
      }

    @Test
    public void shouldReturnFalseAtFirstDisabledCapability()
      throws Exception
      {
        final GenericCapabilityAction<Object> fixture = spy(objectAction);
        final InOrder inOrder = inOrder(fixture);

        doReturn(false).when(fixture).enable(eq(stuff1));

        assertFalse(fixture.enable(createNodes(capabilities)));

        for (final Object capability : capabilities)
          {
            inOrder.verify(fixture).enable(same(capability));

            if (capability == stuff1)
              {
                break;
              }
          }

//        verifyNoMoreInteractions(fixture); // FIXME
      }

    @Test(expected=IllegalArgumentException.class)
    public void shouldThrowExceptionWhenNoCapabilitiesAreSpecified()
      {
        new GenericCapabilityAction<Object>()
          {
            @Override
            protected void performAction (List<Object> capabilities)
              {
              }
          };
      }

    @Test
    public void shouldBeEnabledByDefault()
      {
        assertTrue(objectAction.enable(new Object()));
      }

    @Test
    public void shouldBeInMODE_ANYByDefault()
      {
        assertEquals(CookieAction.MODE_ANY, objectAction.mode());
      }

    @Test
    public void shouldReturnTheSpecifiedCapabilityClasses()
      {
        assertArrayEquals(new Class<?>[]{Object.class}, objectAction.cookieClasses());
        assertArrayEquals(new Class<?>[]{Object.class}, objectAction.getCapabilityClasses());

        assertArrayEquals(new Class<?>[]{Stuff.class}, stuffAction.cookieClasses());
        assertArrayEquals(new Class<?>[]{Stuff.class}, stuffAction.getCapabilityClasses());

        assertArrayEquals(new Class<?>[]{StuffSubClass.class}, subStuffAction.cookieClasses());
        assertArrayEquals(new Class<?>[]{StuffSubClass.class}, subStuffAction.getCapabilityClasses());
      }

    private static Node[] createNodes (@Nonnull final List<Object> capabilities)
      {
        final List<Node> result = new ArrayList<Node>();

        for (final Object capability : capabilities)
          {
            result.add(new AbstractNode(Children.LEAF, Lookups.fixed(capability)));
          }

        return result.toArray(new Node[0]);
      }
  }