/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: EnhancedMultiDataObject.java,v 4e8a394a80a9 2010/03/28 20:51:11 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.loaders;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.io.IOException;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;
import org.openide.util.Lookup;
import org.openide.util.lookup.ProxyLookup;
import org.netbeans.platformx.inject.api.Injector;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.As;
import it.tidalwave.util.As.NotFoundBehaviour;
import it.tidalwave.util.AsException;
import it.tidalwave.netbeans.role.RoleSet;
import it.tidalwave.netbeans.loaders.role.LifeCycleRole;
import it.tidalwave.netbeans.loaders.role.ReplaceProvider;
import it.tidalwave.netbeans.capabilitiesprovider.LookupFactory;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: $
 *
 **********************************************************************************************************************/
public abstract class EnhancedMultiDataObject extends MultiDataObject implements As
  {
    private static final String CLASS = EnhancedMultiDataObject.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    private static final long serialVersionUID = 6098374451101039753L;

    private static int count = 0;

    @CheckForNull
    private transient Lookup lookup;

    @CheckForNull
    private transient Lookup lookup2;

    // FIXME: it is not restored after a deserialization
    // Should be done with a Replace which then delegates to the one create by the ReplaceProvider
    // What about implementing @PrePassivate and @PostActivate?
    @CheckForNull
    private transient RoleSet roleSet;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected EnhancedMultiDataObject (final @Nonnull FileObject fileObject, final @Nonnull MultiFileLoader dataLoader)
      throws DataObjectExistsException
      {
        super(fileObject, dataLoader);
        initializeRoles();
        Injector.getDefault().inject(this, getLookupSafelyInConstructor());
        fireDataObjectCreated();
        count++;
        logger.finest("Total number of EnhancedMultiDataObject instances: %d", count);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public final Object writeReplace()
      {
        final ReplaceProvider replaceProvider = getLookup().lookup(ReplaceProvider.class);
        return (replaceProvider == null) ? super.writeReplace() : replaceProvider.createReplace(this);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public synchronized Lookup getLookup() // FIXME: should be final!
      {
        if (lookup == null)
          {
            lookup = new ProxyLookup(getLookupSafelyInConstructor(), super.getLookup());
          }

        return lookup;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public <T> T as (final @Nonnull Class<T> clazz)
      {
        return as(clazz, As.Defaults.throwAsException(clazz));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public <T> T as (final @Nonnull Class<T> clazz, NotFoundBehaviour<T> notFoundBehaviour)
      {
        final T object = getLookup().lookup(clazz);
        return (object != null) ? object : notFoundBehaviour.run(new AsException(clazz));
      }

    /*******************************************************************************************************************
     *
     * This method creates the specific Lookup for this object and, instead of getLookup(), can be safely invoked from
     * the constructor since it is not overridden. But it can't create a Lookup which includes super Lookup. since
     * when called from the constructor super.getLookup() returns null.
     *
     ******************************************************************************************************************/
    @Nonnull
    private synchronized Lookup getLookupSafelyInConstructor()
      {
        if (lookup2 == null)
          {
            lookup2 = createLookup();
          }

        return lookup2;
      }

    /*******************************************************************************************************************
     *
     * Can be overridden in tests.
     *
     ******************************************************************************************************************/
    @Nonnull
    /* package */ Lookup createLookup()
      {
         return LookupFactory.createLookup(this);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    protected final void handleDelete()
      throws IOException
      {
        logger.fine("handleDelete()");
        super.handleDelete();
        fireDataObjectRemoved();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    protected void finalize()
      {
        count--;
        logger.finest("Total number of EnhancedMultiDataObject instances: %d", count);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void initializeRoles()
      {
        roleSet = new RoleSet();
        roleSet.setInjectedLookup(getLookupSafelyInConstructor());
        roleSet.initialize();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void fireDataObjectCreated()
      {
        logger.fine("fireDataObjectCreated() - %s", this);

        for (final LifeCycleRole role : getLookupSafelyInConstructor().lookupAll(LifeCycleRole.class))
          {
            try
              {
                role.dataObjectCreated(this);
              }
            catch (Throwable t)
              {
                logger.throwing(CLASS, "", t);
              }
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void fireDataObjectRemoved()
      {
        logger.fine("fireDataObjectRemoved() - %s", this);

        for (final LifeCycleRole role : getLookup().lookupAll(LifeCycleRole.class))
          {
            try
              {
                role.dataObjectRemoved(this);
              }
            catch (Throwable t)
              {
                logger.throwing(CLASS, "", t);
              }
          }
      }
  }
