/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: FileFilteredNode.java,v 4e8a394a80a9 2010/03/28 20:51:11 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.filesystem.node.impl;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.io.Serializable;
import java.awt.Image;
import javax.swing.filechooser.FileSystemView;
import org.openide.nodes.Node.Handle;
import org.openide.nodes.FilterNode;
import org.openide.nodes.Node;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataObject;
import org.openide.util.Utilities;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.swing.IconResizer;
import it.tidalwave.netbeans.persistence.TxTask;
import org.openide.filesystems.FileLock;

/***********************************************************************************************************************
 *
 * This node decorator is used to filter out file nodes according to a given
 * criterion; it also binds icons according to the file type.
 *
 * @author  Fabrizio Giudici
 * @version $Id: FileFilteredNode.java,v 4e8a394a80a9 2010/03/28 20:51:11 fabrizio $
 *
 **********************************************************************************************************************/
class FileFilteredNode extends FilterNode 
  {
    private static final String CLASS = FileFilteredNode.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    
    private static final int ICON_SIZE = 16;
        
    private static final Image folderNodeIcon = getIcon("Generic folder.png");
    
    private static final Image openFolderNodeIcon = getIcon("library folder.png");
    
    private static final Image computerIcon = getIcon("Monitor Normal 4to3.png");
    
    private static final Image hddIcon = getIcon("proline black green.png");
    
    private static final Image homeIcon = getIcon("home.png");
    
    private static final Image lockedIcon = getIcon("private folder.png");
    
    private final Node delegateNode;

    private final FileFilter fileFilter;
    
    private Image icon;
    
    private Image openedIcon;
    
    /***************************************************************************
     *
     * This specific Handle manages serialization of the delegate.
     *
     **************************************************************************/
    static class DelegateHandle implements Node.Handle
      {
        private static final long serialVersionUID = 3242335239823432432L;

        private Handle delegateHandle;

        private FileFilter fileFilter;

        public DelegateHandle() // for serialization
          {
          }

        public DelegateHandle (final Node delegateNode, final FileFilter fileFilter)
          {
            if (!(fileFilter instanceof Serializable))
              {
                throw new RuntimeException("If you need getHandle(), the fileFilter must be Serializable");
              }

            delegateHandle = delegateNode.getHandle();
            this.fileFilter = fileFilter;
          }

        public Node getNode()
          throws IOException
          {
            return new FileFilteredNode(delegateHandle.getNode(), fileFilter);
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    static class FileFilteredChildren extends FilterNode.Children 
      {
        private final FileFilter fileFilter;
    
        public FileFilteredChildren (final Node owner, final FileFilter fileFilter) 
          {
            super(owner);
            this.fileFilter = fileFilter;
          }
        
        /***********************************************************************
         *
         *
         **********************************************************************/
        @Override
        protected Node copyNode (final Node original) 
          {
            return new FileFilteredNode(original, fileFilter);
          }
        
        /***********************************************************************
         *
         *
         **********************************************************************/
        @Override
        protected Node[] createNodes (final Node object) 
          {
            final List<Node> result = new ArrayList<Node>();
            
            for (final Node node : super.createNodes(object))
              {
                final DataObject dataObject = node.getLookup().lookup(DataObject.class);  

                if (dataObject != null)
                  {
                    final FileObject fileObject = dataObject.getPrimaryFile();
                    final File file = FileUtil.toFile(fileObject);
                    
                    if (fileFilter.accept(file))
                      {
                        result.add(node);
                      }
                  }
              }
            
            return result.toArray(new Node[0]);
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public FileFilteredNode (final Node delegateNode, final FileFilter fileFilter)
      {
        super(delegateNode, new FileFilteredChildren(delegateNode, fileFilter));
        this.delegateNode = delegateNode;
        this.fileFilter = fileFilter;
        final DataObject dataObject = delegateNode.getLookup().lookup(DataObject.class);
        
        if (dataObject != null)
          {
            final FileObject fileObject = dataObject.getPrimaryFile();
            final File file = FileUtil.toFile(fileObject);
            
            // TODO: setChildren can't be called from this constructor
            // Find some static trickery to pass directly a Children.LEAF in the super()
            // See http://www.nabble.com/Strange-error-with-Children.LEAF-t2496993.html
//            if (!fileObject.isFolder())
//              {
//                setChildren(Children.LEAF);
//              }

            boolean isRoot = fileObject.isRoot();
            final File parentFile = file.getParentFile();                
            
            if ((Utilities.getOperatingSystem() == Utilities.OS_MAC) && (parentFile != null))
              {
                isRoot |= parentFile.getPath().equals("/Volumes");
              }
            
            if (Utilities.isUnix() && (parentFile != null))
              {
                isRoot |= parentFile.getPath().equals("/mnt");
                isRoot |= parentFile.getPath().equals("/media");
              }
            
            if (isRoot)
              {
                icon = openedIcon = hddIcon;
                setSystemName(file);
              }
            
            else if ((Utilities.getOperatingSystem() == Utilities.OS_MAC) && file.getName().toLowerCase().endsWith(".app"))
              {
                setDisplayName(file.getName().substring(0, file.getName().length() - 4));
                // setSystemIcon(file); useless, I hoped it returned the application icon, instead it returns a folder icon
                // FIXME: set a special icon for representing applications
              }
            
            else if (file.getAbsolutePath().equals(System.getProperty("user.home")))
              {
                icon = openedIcon = homeIcon;  
              }
            
            else if (!fileObject.canRead())
              {
//                setChildren(Children.LEAF); see above note
                icon = openedIcon = lockedIcon;  
              }
          }  
      }

    /***************************************************************************
     *
     * {@inheritDoc}
     *
     * This method is called when a file is going to be renamed. Since renaming
     * a folder involves updating the indexes for all children, this method
     * demarcates a transaction for an atomic change of the database.
     *
     **************************************************************************/
    @Override
    public void setName (@Nonnull final String name)
      {
        logger.fine("setName(%s)", name);
        final FileObject fileObject = getLookup().lookup(FileObject.class);

        try
          {
            TxTask.execute(IOException.class, new TxTask<Void, IOException>()
              {
                @Override
                public Void run() throws IOException
                  {
                    final FileLock lock = fileObject.lock();
                    fileObject.rename(lock, name, fileObject.getExt());
                    lock.releaseLock();
                    return null;
                  }
              });
          }
        catch (IOException e)
          {
            throw new RuntimeException(e);
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Override
    public Image getIcon (final int type) 
      {
        if (icon == null)
          {
            icon = (getParentNode() == null) ? computerIcon : folderNodeIcon;
          }  
            
        return icon;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Override
    public Image getOpenedIcon (final int type) 
      {
        if (openedIcon == null)
          {
            openedIcon = (getParentNode() == null) ? computerIcon : openFolderNodeIcon;
          }  
            
        return openedIcon;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Override
    public Handle getHandle()
      {
        return new DelegateHandle(delegateNode, fileFilter);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    private static Image getIcon (final String name)
      {
        return IconResizer.scaledImage(Utilities.loadImage("it/tidalwave/netbeans/filesystem/node/icons/" + name), ICON_SIZE);
      }
    
    /***************************************************************************
     *
     * Sets the name from the system default representation for the given file.
     *
     **************************************************************************/
    private void setSystemName (final File file) 
      {
        //
        // On Mac OS X FileSystemView.getFileSystemView().getSystemDisplayName("/") just returns "/"
        //
        if ((Utilities.getOperatingSystem() == Utilities.OS_MAC) && file.getAbsolutePath().equals("/"))
          {
            setDisplayName(getMacOSXRootVolumeName());
          }

        else
          {
            //
            // Be careful: for instance, hiberfil.sys on Windows throws exception. Moreover, it
            // could happen if repaint() is issued during filesystem operations:
            // maybe a file has been just deleted, and the tree model hasn't been
            // updated yet.
            //
            try
              {
                setDisplayName(FileSystemView.getFileSystemView().getSystemDisplayName(file));
              }
            catch (Exception e)
              {
                logger.warning("setSystemName(%s): %s", file, e);
              }
          }
      }

    /***************************************************************************
     *
     * Sets the icon from the system default representation for the given file.
     *
     **************************************************************************/
//    private void setSystemIcon (final File file)
//      {
//        //
//        // Be careful: for instance, hiberfil.sys on Windows throws exception. Moreover, it
//        // could happen if repaint() is issued during filesystem operations:
//        // maybe a file has been just deleted, and the tree model hasn't been
//        // updated yet.
//        //
//        try
//          {
//            final Icon systemIcon = FileSystemView.getFileSystemView().getSystemIcon(file);
//
//            if (systemIcon != null)
//              {
//                final BufferedImage image = new BufferedImage(systemIcon.getIconWidth(), systemIcon.getIconHeight(), BufferedImage.TYPE_INT_ARGB);
//                final Graphics g = image.createGraphics();
//                systemIcon.paintIcon(null, g, 0, 0);
//                g.dispose();
//                icon = openedIcon = image;
//              }
//          }
//        catch (Exception e)
//          {
//            logger.warning("setSystemIcon(" + file + "): " + e);
//          }
//      }

    /***************************************************************************
     *
     * Finds out the volume name of root in Mac OS X.
     *
     * @return  the root volume name
     *
     **************************************************************************/
    private static String getMacOSXRootVolumeName()
      {
        for (final File file : new File("/Volumes").listFiles())
          {
            try 
              {
                if (file.getCanonicalPath().equals("/"))
                  {
                    return file.getName();                    
                  }
              } 
            catch (IOException e) 
              {
                logger.warning("getMacOSXRootVolumeName(): %s", e);
              }
          }
        
        return "/"; // fallback case
      }
  }
