/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: EnhancedNode.java,v ad650f4b8a8a 2010/04/04 21:30:37 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.nodes;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.awt.Image;
import java.awt.image.BufferedImage;
import javax.swing.Action;
import org.openide.util.Lookup;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.As;
import it.tidalwave.util.AsException;
import it.tidalwave.util.IconProvider;
import it.tidalwave.util.IconProvider.Type;
import it.tidalwave.util.MutableDisplayable;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.util.Removable;
import it.tidalwave.util.Task;
import it.tidalwave.netbeans.aspect.Aspects;
import static it.tidalwave.util.Displayable.Displayable;
import static it.tidalwave.util.IconProvider.Type.*;
import static it.tidalwave.util.MutableDisplayable.MutableDisplayable;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: EnhancedNode.java,v ad650f4b8a8a 2010/04/04 21:30:37 fabrizio $
 *
 **********************************************************************************************************************/
@edu.umd.cs.findbugs.annotations.SuppressWarnings("HE_INHERITS_EQUALS_USE_HASHCODE")
public class EnhancedNode extends AbstractNode
  {
    private static final String CLASS = EnhancedNode.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    private final static BufferedImage EMPTY_ICON = new BufferedImage(16, 16, BufferedImage.TYPE_4BYTE_ABGR);
    
    @Nonnull
    private final Object object;

    private String name;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public EnhancedNode (final @Nonnull Object object,
                         final @Nonnull Children children,
                         final @Nonnull Lookup lookup)
      {
        super(children, lookup);
        this.object = object;

        try
          {
            name = find(Displayable).getDisplayName();
          }
        catch (NotFoundException e)
          {
            name = "???"; // do this first
            logger.severe("No Displayable available for %s", this);
          }

        installDisplayNameListener();
        installIconChangeListener();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public EnhancedNode (final @Nonnull Object object, final @Nonnull Children children)
      {
        this(object, children, Lookup.EMPTY);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public final void setName (final @Nonnull String name)
      {
        this.name = name;
        super.setName(name);

        Aspects.run(this, new Task<Void, RuntimeException>("setName()")
          {
            @Override
            public Void run()
              {
                try
                  {
                    find(MutableDisplayable).setDisplayName(name);
                  }
                catch (NotFoundException e)
                  {
                    // FIXME: do nothing or throw runtime exception?
                  }
                return null;
              }
          });
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public final String getName()
      {
        return name;
      }

//    /*******************************************************************************************************************
//     *
//     * {@inheritDoc}
//     *
//     ******************************************************************************************************************/
//    @Override
//    public final void setDisplayName (final @Nonnull String displayName)
//      {
//        super.setDisplayName(displayName);
//        _setName(displayName);
//      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public final String getDisplayName()
      {
        return name;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Image getIcon (final int type)
      {
        return findIcon(DEFAULT, type);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Image getOpenedIcon (final int type)
      {
        return findIcon(OPEN, type);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public void destroy()
      throws IOException
      {
        final Removable removable = getLookup().lookup(Removable.class);

        if (removable != null)
          {
            try
              {
                Aspects.run(this, new Task<Void, Exception>("destroy()")
                  {
                    @Override
                    public Void run()
                      throws Exception
                      {
                        removable.remove();
                        return null;
                      }
                  });
              }
            catch (IOException e)
              {
                throw e;
              }
            catch (Exception e)
              {
                throw new IOException(e);
              }
          }

        super.destroy();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean canCopy()
      {
        return false;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean canCut()
      {
        return canCopy() && canDestroy();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public boolean canDestroy()
      {
        return getLookup().lookup(Removable.class) != null;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public final boolean canRename()
      {
        try
          {
            find(MutableDisplayable);
            return true;
          }
        catch (NotFoundException e)
          {
            return false;
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @CheckForNull
    public Action getPreferredAction()
      {
        try
          {
            return find(ActionProvider.class).getPreferredAction();
          }
        catch (NotFoundException e)
          {
            return null;
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @CheckForNull
    public Action[] getActions (final boolean context)
      {
        try
          {
            return find(ActionProvider.class).getActions().toArray(new Action[0]);
          }
        catch (NotFoundException e)
          {
            return new Action[0];
          }
      }

//    @Override
//    protected SystemAction[] createActions()
//      {
//        return super.createActions();
//      }
//
//    @Override
//    public SystemAction[] getActions()
//      {
//        return super.getActions();
//      }
//
//    @Override
//    public SystemAction getDefaultAction()
//      {
//        return super.getDefaultAction();
//      }
//
//    @Override
//    public SystemAction[] getContextActions()
//      {
//        return super.getContextActions();
//      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String toString()
      {
        return String.format("%s@%x[name=%s, displayName=%s]",
                             getClass().getSimpleName(), System.identityHashCode(this), getName(), name);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private <T> T find (final @Nonnull Class<T> personality)
      throws NotFoundException
      {
        if (object instanceof As)
          {
            try
              {
                return ((As)object).as(personality);
              }
            catch (AsException e)
              {
                // ok
              }
          }

        final T lookedUp = getLookup().lookup(personality);

        if (lookedUp != null)
          {
            return lookedUp;
          }

        // Check this after as(...) and lookup()
        if (personality.isAssignableFrom(object.getClass()))
          {
            return (T)object;
          }

        throw new NotFoundException();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private Image findIcon (final @Nonnull Type type, final int sizeType)
      {
        final IconProvider iconProvider = getLookup().lookup(IconProvider.class);
        return (iconProvider == null) ? EMPTY_ICON : iconProvider.getIcon(type, 16);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void installDisplayNameListener()
      {
        MutableDisplayable md = null;
        try
          {
            md = find(MutableDisplayable);
            // FIXME: weak listener? Is it possible to replace the MutableDisplayable in the Lookup?
            md.addPropertyChangeListener(new PropertyChangeListener()
              {
                @Override
                public void propertyChange (final @Nonnull PropertyChangeEvent event)
                  {
                    name = (String)event.getNewValue();
                    fireDisplayNameChange((String)event.getOldValue(), name);
                  }
              });
          }
        // FIXME: Currently ElmoSkosConcept doesn't implement properly MutableDisplayable
        // Keeping this block until it's fixed
        catch (AbstractMethodError e) 
          {
            final StackTraceElement[] stackTrace = e.getStackTrace();

            if ((stackTrace != null) && (stackTrace.length > 0)) // FIXME
              {
                logger.warning("BUG (harmless) for %s: %s at %s", md, e, stackTrace[0]);
              }
            else
              {
                logger.warning("BUG (harmless) for %s: %s", md, e);
              }
          }
        // END FIXME
        catch (NotFoundException e)
          {
            // ok, a MutableDisplayable is not mandatory
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void installIconChangeListener()
      {
        final IconProvider iconProvider = getLookup().lookup(IconProvider.class);

        if (iconProvider != null)
          {
            // FIXME: weak listener? Is the IconProvider allowed to change?
            iconProvider.addPropertyChangeListener(new PropertyChangeListener()
              {
                @Override
                public void propertyChange (final @Nonnull PropertyChangeEvent event)
                  {
                    fireIconChange();
                  }
              });
          }
      }
  }
