/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: RetryingChildFactoryExceptionHandler.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.nodes.role;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.WeakHashMap;
import java.text.MessageFormat;
import java.io.IOException;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import javax.swing.Timer;
import lombok.Getter;
import lombok.Setter;
import org.openide.util.RequestProcessor;
import org.openide.util.lookup.Lookups;
import org.openide.util.NbBundle;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.ChildFactory;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.nodes.SafeChildFactory;
import it.tidalwave.netbeans.nodes.role.impl.ErrorNode;
import it.tidalwave.netbeans.nodes.role.impl.Retrier;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: RetryingChildFactoryExceptionHandler.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
public class RetryingChildFactoryExceptionHandler implements ChildFactoryExceptionHandler
  {
    private static final String CLASS = RetryingChildFactoryExceptionHandler.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    @Getter @Setter @Nonnegative
    private int retryPeriod = 10 * 1000;

    @Getter @Setter @Nonnegative
    private int refreshPeriod = 200;

    @Getter @Setter @Nonnegative
    private int iconSize = 16;

    @Getter @Setter @Nonnegative
    private int maxRetryCount = 4;

    @Getter @Setter @Nonnull
    private Color iconColor = new Color(255, 64, 64);

    @Getter @Setter @Nonnull
    private String retryMessage = 
            NbBundle.getMessage(RetryingChildFactoryExceptionHandler.class, "ChildFactoryExceptionHandler.retry");

    @Getter @Setter @Nonnull
    private String errorMessage = 
            NbBundle.getMessage(RetryingChildFactoryExceptionHandler.class, "ChildFactoryExceptionHandler.error");

    @Getter @Setter @Nonnull
    private String errorIconPath =
            NbBundle.getMessage(DefaultChildFactoryExceptionHandler.class, "ChildFactoryExceptionHandler.errorIcon");

    private final WeakHashMap<Node, AtomicInteger> retryCountMap = new WeakHashMap<Node, AtomicInteger>();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    class CountDownNode extends AbstractNode
      {
        private long nextRetryTime;
        
        private String oldDisplayName;

        @CheckForNull
        private Image icon;

        private final Timer timer = new Timer(refreshPeriod, new ActionListener()
          {
            @Override
            public void actionPerformed (final @Nonnull ActionEvent event)
              {
                fireDisplayNameChange(oldDisplayName, getDisplayName());
                fireIconChange();
              }
          });
        
        public CountDownNode (final @Nonnull Throwable t)
          {
            super(Children.LEAF, Lookups.fixed(t));
            nextRetryTime = System.currentTimeMillis() + retryPeriod;
            oldDisplayName = getDisplayName();
            timer.start();
          }

        @Override
        public void destroy()
           throws IOException
          {
            timer.stop();
            super.destroy();
          }

        @Override @Nonnull
        public String getDisplayName()
          {
            final long seconds = 1 + (nextRetryTime - System.currentTimeMillis()) / 1000;
            return MessageFormat.format(retryMessage, seconds);
          }

        @Override @Nonnull
        public Image getIcon (int type)
          {
            if (icon == null)
              {
                icon = new BufferedImage(iconSize, iconSize, BufferedImage.TYPE_4BYTE_ABGR);
              }
            
            final Graphics2D g = (Graphics2D)icon.getGraphics();

//            g.setColor(new Color(0, 0, 0, 0));
            g.setColor(Color.WHITE);
            g.fillRect(0, 0, iconSize, iconSize);
            g.setColor(iconColor);
            g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            final int angle = (int) (360 * (nextRetryTime - System.currentTimeMillis())) / retryPeriod;
            g.fillArc(0, 0, iconSize - 1, iconSize - 1, 90, angle);
            g.drawOval(0, 0, iconSize - 1, iconSize - 1);

            return icon;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void handleException (final @CheckForNull Node node,
                                 final @Nonnull ChildFactory childFactory,
                                 final @Nonnull Throwable throwable)
      {
        logger.fine("handleException(%s, %s, %s)", node, childFactory, throwable);
        logger.warning("While populating children nodes: ", throwable);
        logger.throwing(CLASS, "handleException()", throwable);

        if (node != null)
          {
            final Children children = node.getChildren();
            children.remove(children.getNodes());
            final Retrier retrier = new Retrier(children, (SafeChildFactory<?>)childFactory);
            AtomicInteger retryCount = retryCountMap.get(node);

            if (retryCount == null)
              {
                retryCountMap.put(node, retryCount = new AtomicInteger(maxRetryCount));
              }

            if (retryCount.getAndDecrement() <= 0)
              {
                logger.finer(">>>> no more retries, setting error node");
                retryCountMap.remove(node);
                children.add(new Node[]{new ErrorNode(errorMessage, errorIconPath, throwable, retrier)});
              }
            else
              {
                logger.finest(">>>> available retries: %s", retryCount);
                children.add(new Node[]{new CountDownNode(throwable)});

                RequestProcessor.getDefault().post(new Runnable()
                  {
                    @Override
                    public void run()
                      {
                        try
                          {
                            Thread.sleep(retryPeriod);
                            retrier.retry();
                          }
                        catch (InterruptedException e)
                          {
                          }
                      }
                  });
              }
          }
      }
  }
