/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: EnhancedNodeTest.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.nodes;

import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import it.tidalwave.util.As;
import it.tidalwave.util.AsException;
import it.tidalwave.util.Displayable;
import it.tidalwave.util.MutableDisplayable;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;
import static org.hamcrest.CoreMatchers.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: EnhancedNodeTest.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
public class EnhancedNodeTest
  {
    @Test
    public void mustWorkWithClassesImplementingDisplayable()
      {
        final Displayable displayable = mock(Displayable.class);
        when(displayable.getDisplayName()).thenReturn("THE NAME 1");
        final Node fixture = new EnhancedNode(displayable, Children.LEAF);
        assertThat(fixture.getName(), is("THE NAME 1"));
        assertThat(fixture.getDisplayName(), is("THE NAME 1"));
        assertThat(fixture.canRename(), is(false));
      }

    @Test
    public void mustWorkWithClassesImplementingAsDisplayable()
      {
        final Displayable displayable = mock(Displayable.class);
        when(displayable.getDisplayName()).thenReturn("THE NAME 2");
        final As as = mock(As.class);
        when(as.as(eq(Displayable.class))).thenReturn(displayable);
        when(as.as(eq(MutableDisplayable.class))).thenThrow(new AsException(MutableDisplayable.class));
        final Node fixture = new EnhancedNode(as, Children.LEAF);
        assertThat(fixture.getName(), is("THE NAME 2"));
        assertThat(fixture.getDisplayName(), is("THE NAME 2"));
        assertThat(fixture.canRename(), is(false));
      }

    @Test
    public void mustRenderQuestionMarksWithClassesImplementingAsButNotADisplayable()
      {
        final As as = mock(As.class);
        when(as.as(eq(Displayable.class))).thenThrow(new AsException(Displayable.class));
        when(as.as(eq(MutableDisplayable.class))).thenThrow(new AsException(MutableDisplayable.class));
        final Node fixture = new EnhancedNode(as, Children.LEAF);
        assertThat(fixture.getName(), is("???"));
        assertThat(fixture.getDisplayName(), is("???"));
        assertThat(fixture.canRename(), is(false));
      }

    @Test
    public void mustRenderQuestionMarksWithClassesNotImplementingAs()
      {
        final Object object = mock(Object.class);
        final Node fixture = new EnhancedNode(object, Children.LEAF);
        assertThat(fixture.getName(), is("???"));
        assertThat(fixture.getDisplayName(), is("???"));
        assertThat(fixture.canRename(), is(false));
      }

    @Test
    public void mustWorkWithClassesImplementingMutableDisplayable()
      {
        final MutableDisplayable mutableDisplayable = mock(MutableDisplayable.class);
        when(mutableDisplayable.getDisplayName()).thenReturn("THE NAME 1");
        final Node fixture = new EnhancedNode(mutableDisplayable, Children.LEAF);
        assertThat(fixture.getName(), is("THE NAME 1"));
        assertThat(fixture.getDisplayName(), is("THE NAME 1"));
        assertThat(fixture.canRename(), is(true));
        fixture.setName("NEW NAME");
        verify(mutableDisplayable).setDisplayName(eq("NEW NAME"));
        assertThat(fixture.getName(), is("NEW NAME"));
        assertThat(fixture.getDisplayName(), is("NEW NAME"));
      }

    @Test
    public void mustWorkWithClassesImplementingAsMutableDisplayable()
      {
        final MutableDisplayable mutableDisplayable = mock(MutableDisplayable.class);
        when(mutableDisplayable.getDisplayName()).thenReturn("THE NAME 2");
        final As as = mock(As.class);
        when(as.as(eq(Displayable.class))).thenReturn(mutableDisplayable);
        when(as.as(eq(MutableDisplayable.class))).thenReturn(mutableDisplayable);
        final Node fixture = new EnhancedNode(as, Children.LEAF);
        assertThat(fixture.getName(), is("THE NAME 2"));
        assertThat(fixture.getDisplayName(), is("THE NAME 2"));
        assertThat(fixture.canRename(), is(true));
        fixture.setName("NEW NAME");
        verify(mutableDisplayable).setDisplayName(eq("NEW NAME"));
        assertThat(fixture.getName(), is("NEW NAME"));
        assertThat(fixture.getDisplayName(), is("NEW NAME"));
      }

    @Test
    public void mustPassChildrenToSuperClass()
      {
        final Object as = mock(Object.class);
        final Children children = mock(Children.class);
        final Node fixture = new EnhancedNode(as, children);
        assertThat(fixture.getChildren(), is(sameInstance(children)));
      }

    @Test
    public void mustPassChildrenAndLookupToSuperClass()
      {
        final Object as = mock(Object.class);
        final Children children = mock(Children.class);
        final Lookup lookup = Lookups.fixed();
        final Node fixture = new EnhancedNode(as, children, lookup);
        assertThat(fixture.getChildren(), is(sameInstance(children)));
        assertThat(fixture.getLookup(), is(sameInstance(lookup)));
      }
  }