/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: RoleSetTest.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.role;

import java.util.Arrays;
import java.util.Collection;
import javax.annotation.Nonnull;
import javax.annotation.PostConstruct;
import javax.inject.Inject;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;
import it.tidalwave.netbeans.util.test.MockLookup;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;
import static org.hamcrest.CoreMatchers.*;
import static org.junit.matchers.JUnitMatchers.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: RoleSetTest.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
public class RoleSetTest
  {
    public static class Role1
      {
        @Inject
        public Lookup lookup;

        @PostConstruct
        @SuppressWarnings("PMD.UnusedPrivateMethod")
        public void initialize()
          {
          }
      }

    public static class Role2
      {
        @Inject
        public Lookup lookup;

        @Inject
        public Role1 role1;

        @PostConstruct
        @SuppressWarnings("PMD.UnusedPrivateMethod")
        public void initialize()
          {
          }
      }

    private RoleSet roleSet;

    private Role1 role1;
    private Role2 role2;
    private Object role3;
    private Object role4;

    private Role2 role5;
    private Object role6;

    @BeforeClass
    public static void setupGlobalLookup()
      {
        MockLookup.setInstances("G1", "G2");
      }

    @Before
    public void setUp()
      {
        roleSet = new RoleSet();
        role1 = spy(new Role1());
        role2 = spy(new Role2());
        role3 = new Object();
        role4 = new Object();
        role5 = spy(new Role2());
        role6 = new Object();
      }

    @Test
    public void lookupShouldContainAllStaticRoles()
      {
        roleSet.setStaticRoles(Arrays.asList(role1, role2, role3, role4));
        roleSet.initialize();
        assertLookupContains(roleSet.getLookup(), role1, role2, role3, role4);
      }

    @Test
    public void lookupShouldContainAddedRoles()
      {
        roleSet.addRole(role1);
        roleSet.addRole(role2);
        roleSet.initialize();
        roleSet.addRole(role3);
        roleSet.addRole(role4);
        assertLookupContains(roleSet.getLookup(), role1, role2, role3, role4);
      }

    @Test
    public void lookupShouldNotContainRemovedRoles()
      {
        roleSet.addRole(role1);
        roleSet.addRole(role2);
        roleSet.initialize();
        assertLookupContains(roleSet.getLookup(), role1, role2);
        roleSet.removeRole(role1);
        roleSet.removeRole(role2);
        assertLookupContains(roleSet.getLookup());
      }

    @Test
    public void shouldNotRemoveStaticRoles()
      {
        roleSet.setStaticRoles(Arrays.asList(role1, role2, role3, role4));
        roleSet.addRole(role5);
        roleSet.initialize();
        roleSet.addRole(role6);
        assertLookupContains(roleSet.getLookup(), role1, role2, role3, role4, role5, role6);
        roleSet.removeRole(role1);
        roleSet.removeRole(role2);
        roleSet.removeRole(role3);
        roleSet.removeRole(role4);
        roleSet.removeRole(role5);
        roleSet.removeRole(role6);
        assertLookupContains(roleSet.getLookup(), role1, role2, role3, role4);
      }

    @Test
    public void shouldInitializeRoles()
      {
        final Lookup injectedLookup = Lookups.fixed("a", "b", "c", "d");
        roleSet.setInjectedLookup(new ProxyLookup(roleSet.getLookup(), injectedLookup)); // typical use
        roleSet.initialize();
        
        roleSet.addRole(role1);
        roleSet.addRole(role2);

        verify(role1).initialize();
        verifyNoMoreInteractions(role1);
        assertLookupContains(role1.lookup, "a", "b", "c", "d", "G1", "G2");

        verify(role2).initialize();
        verifyNoMoreInteractions(role2);
        assertLookupContains(role2.lookup, "a", "b", "c", "d", "G1", "G2");
        assertThat(role2.role1, is(sameInstance(role1)));
      }

    @Test
    public void shouldNotInitializeRolesTooEarly()
      {
        roleSet.addRole(role1);
        verifyZeroInteractions(role1);
        assertThat(role1.lookup, is(nullValue()));

        roleSet.addRole(role2);
        verifyZeroInteractions(role2);
        assertThat(role2.lookup, is(nullValue()));

        final Lookup injectedLookup = Lookups.fixed("a", "b", "c", "d");
        roleSet.setInjectedLookup(new ProxyLookup(roleSet.getLookup(), injectedLookup)); // typical use
        roleSet.initialize();

        verify(role1).initialize();
        verifyNoMoreInteractions(role1);
        assertLookupContains(role1.lookup, "a", "b", "c", "d");

        verify(role2).initialize();
        verifyNoMoreInteractions(role2);
        assertLookupContains(role2.lookup, "a", "b", "c", "d");
        assertThat(role2.role1, is(sameInstance(role1)));
      }

    private static void assertLookupContains (final @Nonnull Lookup lookup, final Object ... objects)
      {
        final Collection<Object> contents = (Collection<Object>)lookup.lookupAll(Object.class);
//        assertThat(contents.toString(), contents.size(), is(objects.length));

        for (final Object object : objects)
          {
            assertThat(contents, hasItem(object));
          }
//        if (objects.length > 0)
//          {
//            assertThat(contents, JUnitMatchers.<Object>hasItems(objects));
//          }
      }
  }
