/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: BeanFactoryTest.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.role.util;

import java.awt.Component;
import java.awt.EventQueue;
import java.util.Arrays;
import java.util.Collections;
import org.openide.filesystems.FileObject;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;
import static org.hamcrest.CoreMatchers.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: BeanFactoryTest.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
public class BeanFactoryTest
  {
    private FileObject fileObject;

    public static class Bean1
      {
        protected String property1;
        protected int property2;
        protected String readOnlyProperty;

        public String getProperty1()
          {
            return property1;
          }

        public void setProperty1 (String property1)
          {
            this.property1 = property1;
          }

        public int getProperty2()
          {
            return property2;
          }

        public void setProperty2 (int property2)
          {
            this.property2 = property2;
          }

        public String getReadOnlyProperty()
          {
            return readOnlyProperty;
          }
      }

    public static class Bean2 extends Component
      {
        public boolean createdInEDTThread;

        public Bean2()
          {
            createdInEDTThread = EventQueue.isDispatchThread();
          }
      }
    
    @Before
    public void setUp()
      {
        fileObject = mock(FileObject.class);
        when(fileObject.getAttribute(eq("instanceCreate"))).thenReturn("foobar");
        when(fileObject.getAttribute(eq("class"))).thenReturn(Bean1.class.getName());
        when(fileObject.getAttribute(eq("property1"))).thenReturn("value1");
        when(fileObject.getAttribute(eq("property2"))).thenReturn(2);
        when(fileObject.getAttributes()).thenReturn(Collections.enumeration(Arrays.asList("class", "property1", "property2")));
      }

    @Test
    public void shouldProperlyInstantiateBean1()
      throws Exception
      {
        final Bean1 bean1 = (Bean1)BeanFactory.createInstance(fileObject);
        assertThat(bean1, is(not(nullValue())));
        assertThat(bean1.getProperty1(), is("value1"));
        assertThat(bean1.getProperty2(), is(2));
      }

    @Test(expected=IllegalArgumentException.class)
    public void shouldThrowIllegalArgumentExceptionWhenMissingClassAttribute()
      throws Exception
      {
        when(fileObject.getAttribute(eq("class"))).thenReturn(null);
        when(fileObject.getAttributes()).thenReturn(Collections.enumeration(Arrays.asList("property1", "property2")));
        BeanFactory.createInstance(fileObject);
      }

    @Test(expected=UnsupportedOperationException.class)
    public void shouldThrowUnsupportedOperationExceptionWhenTryingToSetAReadOnlyProperty()
      throws Exception
      {
        when(fileObject.getAttribute(eq("readOnlyProperty"))).thenReturn("value");
        when(fileObject.getAttributes()).thenReturn(Collections.enumeration(Arrays.asList("class", "property1", "property2", "readOnlyProperty")));
        BeanFactory.createInstance(fileObject);
      }

    @Test(expected=UnsupportedOperationException.class)
    public void shouldThrowUnsupportedOperationExceptionWhenTryingToSetAPropertyWhichDoesntExist()
      throws Exception
      {
        when(fileObject.getAttribute(eq("notExistingProperty"))).thenReturn("value");
        when(fileObject.getAttributes()).thenReturn(Collections.enumeration(Arrays.asList("class", "property1", "property2", "notExistingProperty")));
        BeanFactory.createInstance(fileObject);
      }

    @Test
    public void shouldInstantiateSwingComponentsInTheEDTThreadWhenCalledFromTheAnyThread()
      throws Exception
      {
        final Bean2 bean = (Bean2)BeanFactory.instantiate(Bean2.class);
        assertThat(bean, is(notNullValue()));
        assertThat(bean.createdInEDTThread, is(true));
      }
    
    @Test
    public void shouldInstantiateSwingComponentsInTheEDTThread2FromTheEDTThread()
      throws Exception
      {
        final Bean2[] bean = new Bean2[1];
        
        EventQueue.invokeAndWait(new Runnable() 
          {
            @Override
            public void run() 
              {
                try
                  {
                    bean[0] = (Bean2)BeanFactory.instantiate(Bean2.class);
                  }
                catch (Exception e)
                  {
                    e.printStackTrace();
                  }
              }
          });
          
        assertThat(bean[0], is(notNullValue()));
        assertThat(bean[0].createdInEDTThread, is(true));
      }
 }