/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: Aspect.java,v ad650f4b8a8a 2010/04/04 21:30:37 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.aspect;

import javax.annotation.Nonnull;
import it.tidalwave.util.Task;
import it.tidalwave.util.logging.Logger;

/***********************************************************************************************************************
 *
 * This class allows to implement an aspect that must be run across a task.
 * 
 * <pre>
 * public class MyAspect extends Aspect
 *   {
 *     &#64;Override
 *     protected void runBefore()
 *       {
 *         // things to do before the task
 *       }
 *
 *     &#64;Override
 *     protected void runAfter()
 *       {
 *         // things to do after the task, in case of no exceptions
 *       }
 *
 *     &#64;Override
 *     protected void runFinally()
 *       {
 *         // things to do after the task, always
 *       }
 *
 *     &#64;Override
 *     protected void handleException (final @Nonnull Throwable t)
 *       {
 *         // things to do in case of exception
 *       }
 *   }
 * </pre>
 *
 * It is possible to run a task with the aspect with the following code:
 * 
 * <pre>
 * MyAspect aspect = new MyAspect();
 * aspect.run(new Task&lt;String, NotFoundException&gt;()
 *   {
 *     &#64;Override
 *     public String run()
 *       throws NotFoundException
 *       {
 *         // do something
 *         return "something";
 *       }
 *   });
 * </pre>
 * 
 * @author Fabrizio Giudici
 * @version $Id: Aspect.java,v ad650f4b8a8a 2010/04/04 21:30:37 fabrizio $
 * @experimental
 *
 **********************************************************************************************************************/
public abstract class Aspect
  {
    private static final String CLASS = Aspect.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    /*******************************************************************************************************************
     *
     * A special {@link Aspect} that does nothing.
     *
     ******************************************************************************************************************/
    public final static Aspect VOID = new Aspect()
      {
        @Override @Nonnull
        public String toString()
          {
            return "Aspect.VOID";
          }
      };

    /*******************************************************************************************************************
     *
     * Runs the given {@link Task} applying the aspect methods.
     *
     * @param  task  the {@code Task}
     * @return       the return object from the {@code Task}
     *
     ******************************************************************************************************************/
    final public <T, E extends Throwable> T run (final @Nonnull Task<T, E> task)
      throws E
      {
        logger.fine("run(%s) with %s", task, this);

        try
          {
            runBefore();
            final T result = task.run();
            runAfter();
            return result;
          }
        catch (Throwable t)
          {
            logger.throwing(CLASS, "run()", t);
            handleException(t);

            try
              {
                throw (E)t;
              }
            catch (ClassCastException e)
              {
                throw new RuntimeException(e);
              }
          }
        finally
          {
            runFinally();
          }
      }

    /*******************************************************************************************************************
     *
     * This method is invoked before the {@link Task}.
     *
     ******************************************************************************************************************/
    protected void runBefore()
      {
      }

    /*******************************************************************************************************************
     *
     * This method is invoked after the {@link Task} if no exception is thrown.
     *
     ******************************************************************************************************************/
    protected void runAfter()
      {
      }

    /*******************************************************************************************************************
     *
     * This method is always invoked after the {@link Task}.
     *
     ******************************************************************************************************************/
    protected void runFinally()
      {
      }

    /*******************************************************************************************************************
     *
     * This method is invoked when the {@link Task} throws an exception.
     *
     * @param  t  the thrown exception
     *
     ******************************************************************************************************************/
    protected void handleException (final @Nonnull Throwable t)
      {
      }
  }
