/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: EnhancedTopComponent.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.windows;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.swing.Action;
import java.awt.BorderLayout;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;
import org.openide.filesystems.FileObject;
import org.openide.windows.TopComponent;
import org.netbeans.platformx.inject.api.Injector;
import it.tidalwave.util.As;
import it.tidalwave.util.As.NotFoundBehaviour;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.role.RoleSet;
import it.tidalwave.netbeans.windows.role.GuiBuilder;
import it.tidalwave.netbeans.windows.role.TopComponentRole;
import it.tidalwave.netbeans.windows.impl.action.ToggleComponentAction;

/***********************************************************************************************************************
 *
 * This class enhances a {@link TopComponent} with the capability of plugging in {@link role}s.
 *
 * @author  Fabrizio Giudici
 * @version $Id: EnhancedTopComponent.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
public abstract class EnhancedTopComponent extends TopComponent implements As
  {
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    interface RoleRunner
      {
        public void run (final @Nonnull TopComponentRole role);
      }

    private final static String CLASS = EnhancedTopComponent.class.getName();
    private final static Logger logger = Logger.getLogger(CLASS);

    @Nonnull
    private final Lookup lookup;

    private final RoleSet roleSet = new RoleSet();

    private boolean rolesInstalled = false;

    @CheckForNull
    private final String id;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public EnhancedTopComponent()
      {
        this(null);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public EnhancedTopComponent (@CheckForNull final String id)
      {
        this.id = id;
        lookup = new ProxyLookup(roleSet.getLookup(),
                                 Lookups.fixed(this), // needed by roles
                                 super.getLookup());
        setLayout(new BorderLayout());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override  @Nonnull
    public Lookup getLookup()
      {
        return lookup;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void addRole (@Nonnull final Object role)
      {
        roleSet.addRole(role);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void removerole (@Nonnull final Object role)
      {
        roleSet.removeRole(role);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public <T> T as (final @Nonnull Class<T> clazz)
      {
        return as(clazz, As.Defaults.throwAsException(clazz));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public <T> T as (final @Nonnull Class<T> clazz,
                     final @Nonnull NotFoundBehaviour<T> notFoundBehaviour)
      {
        final T object = getLookup().lookup(clazz);
        return (object != null) ? object : notFoundBehaviour.run(null);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public abstract String preferredID(); // FIXME: for GUIBuilder

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public static Action toggleAction (@Nonnull final FileObject fileObject)
      {
        return new ToggleComponentAction(fileObject);
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    protected void componentActivated()
      {
        super.componentActivated();
        
        runRoles(new RoleRunner()
          {
            @Override
            public void run (final @Nonnull TopComponentRole role)
              {
                role.notifyActivated();
              }
          });
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    protected void componentDeactivated()
      {
        super.componentDeactivated();
        runRoles(new RoleRunner()
          {
            @Override
            public void run (final @Nonnull TopComponentRole role)
              {
                role.notifyDeactivated();
              }
          });
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @edu.umd.cs.findbugs.annotations.SuppressWarnings("SIC_INNER_SHOULD_BE_STATIC_ANON")
    protected void componentClosed()
      {
        super.componentClosed();
        runRoles(new RoleRunner()
          {
            @Override
            public void run (final @Nonnull TopComponentRole role)
              {
                role.notifyClosed();
              }
          });
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @edu.umd.cs.findbugs.annotations.SuppressWarnings("SIC_INNER_SHOULD_BE_STATIC_ANON")
    protected void componentOpened()
      {
        super.componentOpened();
        runRoles(new RoleRunner()
          {
            @Override
            public void run (final @Nonnull TopComponentRole role)
              {
                role.notifyOpened();
              }
          });
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @edu.umd.cs.findbugs.annotations.SuppressWarnings("SIC_INNER_SHOULD_BE_STATIC_ANON")
    protected void componentHidden()
      {
        super.componentHidden();
        runRoles(new RoleRunner()
          {
            @Override
            public void run (final @Nonnull TopComponentRole role)
              {
                role.notifyHidden();
              }
          });
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    @edu.umd.cs.findbugs.annotations.SuppressWarnings("SIC_INNER_SHOULD_BE_STATIC_ANON")
    protected void componentShowing()
      {
        super.componentShowing();
        runRoles(new RoleRunner()
          {
            @Override
            public void run (final @Nonnull TopComponentRole role)
              {
                role.notifyShowing();
              }
          });
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    private void runRoles (final @Nonnull RoleRunner roleRunner)
      {
        //
        // Makes sure that roles are injected and initialized..The reason for which we're doing it now and not
        // in the constructor is because we want to make available the actual Lookup} that could have been enhanced
        // by a subclass, e.g. by overriding the getLookup() method.
        //
        if (!rolesInstalled)
          {
            installRoles();
            Injector.getDefault().inject(this, getLookup());
            rolesInstalled = true;
          }
        
        assert rolesInstalled : "roles not initialized";
        
        for (final TopComponentRole role : getLookup().lookupAll(TopComponentRole.class))
          {
            roleRunner.run(role);
          }
      }

    /*******************************************************************************************************************
     *
     * FIXME: make it private
     *
     ******************************************************************************************************************/
    public void installRoles()
      {
        final String path = "/Roles/" + ((id != null) ? id : preferredID());
        logger.finer(">>>> looking up default roles from path %s...", path);

        roleSet.setInjectedLookup(getLookup()); // a subclass might have enhanced it
        roleSet.setStaticRoles(Lookups.forPath(path).lookupAll(Object.class));
        roleSet.initialize();

        final GuiBuilder guiBuilder = roleSet.getLookup().lookup(GuiBuilder.class);

        if (guiBuilder != null)
          {
            add(guiBuilder.createGui(), BorderLayout.CENTER);
          }
      }
  }
