/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: AutomatedTestRunner.java,v 8f602be640c5 2010/03/28 22:35:05 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.automatedtest;

import it.tidalwave.util.logging.Logger;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import javax.swing.SwingUtilities;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: AutomatedTestRunner.java,v 8f602be640c5 2010/03/28 22:35:05 fabrizio $
 *
 **********************************************************************************************************************/
public final class AutomatedTestRunner
  {
    private static final String CLASS = AutomatedTestRunner.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    
    private final AutomatedTestStatusLine statusLine = AutomatedTestStatusLine.getInstance();
            
    /***************************************************************************
     *
     *
     **************************************************************************/
    public void runTest (final AutomatedTest test, final boolean stopOnBrokenTest)
      throws TestFailedException
      {
        logger.info("runTest(%s)", test);
        assertFalse("Cannot run in the AWT Thread", SwingUtilities.isEventDispatchThread());

        try
          {
            statusLine.setTestName("");
            statusLine.setVisible(true);

            if (test instanceof AutomatedTestSuite)
              {
                for (final AutomatedTest childTest : (AutomatedTestSuite)test)
                  {
                    try
                      {
                        runTest(childTest, stopOnBrokenTest);
                      }
                    catch (TestFailedException e)
                      {
                        if (stopOnBrokenTest)
                          {
                            throw e;
                          }
                      }
                  }
              }

            else
              {
                runSingleTest(test);
              }
          }
        finally
          {
            statusLine.setVisible(false);
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    private void runSingleTest (final AutomatedTest test)
      throws TestFailedException
      {
        logger.info("TEST: running %s", test.getName());
        
        try
          {
            test.setStatus(AutomatedTest.Status.RUNNING);
            statusLine.setTestName(test.getName());
            runBefore(test);
            runTestMethods(test);
            test.setStatus(AutomatedTest.Status.PASSED);
            logger.info("TEST: passed %s", test.getName());
          }
        catch (Throwable e)
          {
            logger.info("TEST: failed %s", test.getName());
            logger.throwing(CLASS, "runSingleTest()", e);
            test.setStatus(AutomatedTest.Status.FAILED);
            throw new TestFailedException(e);
          }
        finally
          {
            try
              {
                runAfter(test);
              }
            catch (Exception e)
              {
                logger.throwing(CLASS, "runSingleTest()", e);
              }
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    private void runBefore (final AutomatedTest test)
      throws SecurityException, IllegalArgumentException, IllegalAccessException, InvocationTargetException 
      {
        statusLine.setOperationName("preparing...");
        
        for (final Method method : test.getClass().getMethods())
          {
            final Before before = method.getAnnotation(Before.class);
            
            if (before != null)
              {
                method.invoke(test);
              }
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    private void runAfter (final AutomatedTest test)
      throws SecurityException, IllegalArgumentException, IllegalAccessException, InvocationTargetException
      {
        statusLine.setOperationName("cleaning up...");

        for (final Method method : test.getClass().getMethods())
          {
            final After after = method.getAnnotation(After.class);

            if (after != null)
              {
                method.invoke(test);
              }
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    private void runTestMethods (final AutomatedTest test)
      throws TestFailedException
      {
        final Method[] methods = test.getClass().getMethods();        
//        logger.finest(">> Methods: " + Arrays.asList(methods));
        
        for (final Method method : methods)
          {
            final Test testMethod = method.getAnnotation(Test.class);

            if (testMethod != null)
              {
                logger.fine("Calling method: " + method.getName());
                statusLine.setOperationName(method.getName());

                try
                  {
                    method.invoke(test);
                  }
                catch (IllegalAccessException e)
                  {
                    throw new TestFailedException(e);
                  }
                catch (IllegalArgumentException e)
                  {
                    throw new TestFailedException(e);
                  }
                catch (InvocationTargetException e)
                  {
                    throw new TestFailedException(e);
                  }
              }
          }
      }
  }
