/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: Utils.java,v 8f602be640c5 2010/03/28 22:35:05 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.automatedtest;

import it.tidalwave.netbeans.dialog.DialogRunner;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.rendezvous.RendezVous;
import it.tidalwave.netbeans.rendezvous.RendezVous.Tag;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.workspacemanager.WorkspaceManager;
import java.lang.reflect.InvocationTargetException;
import java.util.Date;
import java.beans.PropertyVetoException;
import java.io.File;
import java.awt.Robot;
import java.awt.Component;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.ListCellRenderer;
import java.awt.AWTException;
import java.awt.Dialog;
import java.awt.EventQueue;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.swing.JComponent;
import javax.swing.JSlider;
import javax.swing.JToggleButton;
import javax.swing.SwingUtilities;
import javax.swing.text.JTextComponent;
import org.openide.explorer.ExplorerManager;
import org.openide.nodes.Node;
import org.openide.windows.TopComponent;
import org.openide.cookies.InstanceCookie;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.Repository;
import org.openide.loaders.DataObject;
import org.openide.util.RequestProcessor;
import org.openide.util.actions.SystemAction;
import static org.junit.Assert.*;


abstract class AWTExecutor<T>
  {
    private Throwable t;
    private T result;

    public T execute (@Nullable final Tag tag)
      throws Exception
      {
        EventQueue.invokeAndWait(new Runnable()
          {
            public void run()
              {
                try
                  {
                    Tag.set(tag);
                    result = process();
                  }
                catch (Throwable t)
                  {
                    AWTExecutor.this.t = t;
                  }
              }
          });

        EventQueue.invokeAndWait(new Runnable()
          {
            public void run()
              {
                Tag.set(null);
              }
          });

        if (t != null)
          {
            if ((t instanceof Exception))
              {
                throw (Exception)t;
              }

            throw (Error)t;
          }

        return result;
      }

    protected abstract T process()
      throws Exception;
  }

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: Utils.java,v 8f602be640c5 2010/03/28 22:35:05 fabrizio $
 *
 **********************************************************************************************************************/
public final class Utils  
  {
    private static final String CLASS = Utils.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    
    private static Robot robot;
    
    static
      {
        try 
          {
            robot = new Robot();
          } 
        catch (AWTException e) 
          {
            logger.throwing(CLASS, "static", e);
            logger.severe("Can't initialize robot: %s", e);
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    private Utils() 
      {
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static void waitInAWT (@Nonnegative final long timeout,
                                  @Nonnegative final long pollDelay,
                                  @Nonnull final Function function)
      throws Exception
      {
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        final long time = System.currentTimeMillis();

        while ((System.currentTimeMillis() - time) <= timeout)
          {
            final boolean value = new AWTExecutor<Boolean>()
              {
                @Override
                protected Boolean process()
                  throws Exception
                  {
                    return function.evaluate();
                  }
              }.execute(null);

            if (value)
              {
                return;
              }

            Thread.sleep(pollDelay);
          }

        throw new InterruptedException("Timeout expired");
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static void waitFor (@Nonnegative final long timeout,
                                @Nonnegative final long pollDelay,
                                @Nonnull final Function function)
      throws Exception
      {
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        final long time = System.currentTimeMillis();

        while ((System.currentTimeMillis() - time) <= timeout)
          {
            final boolean value = function.evaluate();

            if (value)
              {
                return;
              }

            Thread.sleep(pollDelay);
          }

        throw new InterruptedException("Timeout expired");
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static void waitUntilVisible (final Component component)
      {
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        while (!component.isVisible())
          {
            delay(100);  
          }
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Tag selectNode (final ExplorerManager.Provider explorerManagerProvider, final Node node)
      throws Exception
      {
        logger.info("==================== selectNode(%s, %s)", explorerManagerProvider, node);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        final Tag tag = new Tag();
        
        new AWTExecutor<Void>()
          {
            @Override
            protected Void process()
              throws PropertyVetoException
              {
                final Node[] nodes = (node != null) ? new Node[]{node} : new Node[0];
                explorerManagerProvider.getExplorerManager().setSelectedNodes(nodes);
                return null;
              }
          }.execute(tag);

        return tag;
      }
     
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static void delay (@Nonnegative final long time)
      {
        logger.info("==================== delay(%d)", time);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        try 
          {
            Thread.sleep(time);
          } 
        catch (InterruptedException ex) 
          {
            ex.printStackTrace();
          }
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static void activate (@Nonnull final TopComponent topComponent)
      throws Exception
      { 
        logger.info("==================== activate(%s)", topComponent);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        new AWTExecutor<Void>()
          {
            @Override
            protected Void process() throws Exception
              {
                if (!topComponent.isOpened())
                  {
                    topComponent.open();
                  }

                topComponent.requestActive();
                return null;
              }
        }.execute(null);
//        assertActivated(topComponent.getClass().getName());
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static void setVisible (@Nonnull final TopComponent topComponent, final boolean visible)
      throws Exception
      {
        logger.info("==================== setVisible(%s, %s)", topComponent, visible);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        new AWTExecutor<Void>()
          {
            @Override
            protected Void process()
              {
                if (visible)
                  {
                    if (!topComponent.isOpened())
                      {
                        topComponent.open();
                        topComponent.requestActive();
                      }

                    else if (!topComponent.isVisible())
                      {
                        topComponent.requestActive();
                      }
                  }
                else
                  {
                    topComponent.close();
                  }

                return null;
              }
          }.execute(null);

        waitUntilVisible(topComponent);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static TopComponent assertActivated (final String className)
      throws Exception
      {
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        return new AWTExecutor<TopComponent>()
          {
            @Override
            @Nonnull
            protected TopComponent process()
              {
                final TopComponent activeComponent = TopComponent.getRegistry().getActivated();
                assertNotNull(activeComponent);
                assertEquals(className, activeComponent.getClass().getName());
                return activeComponent;
              }
          }.execute(null);
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static void assertOpened (final String className)
      throws Exception
      {
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        new AWTExecutor<Void>()
          {
            @Override
            protected Void process() throws Exception
              {
                for (TopComponent opened : TopComponent.getRegistry().getOpened())
                  {
                    if (opened.getClass().getName().equals(className))
                      {
                        return null;
                      }
                  }

                fail("Component is not opened: " + className);
                return null; // never reached
              }
          }.execute(null);
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
//    private static void performAction (final String actionClassName)
//      {
//        try
//          {
//            final Class<? extends SystemAction> actionClass = (Class<? extends SystemAction>)Class.forName(actionClassName);
//            performAction(actionClass);
//          }
//        catch (Exception e)
//          {
//            throw new AssertionError(e);
//          }
//      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static void performAction (final Class<? extends SystemAction> actionClass)
      {
        logger.info("==================== performAction(%s)", actionClass);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        try 
          {
            performAction(actionClass.newInstance());
          } 
        catch (Exception e) 
          {
            throw new AssertionError(e);
          }
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Tag performAction (final SystemAction action)
      throws Exception
      {
        logger.info("==================== performAction(%s)", action);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        final Tag tag = new Tag();
        new AWTExecutor<Void>()
          {
            @Override
            protected Void process()
              {
                action.actionPerformed(null); // FIXME
                return null;
              }
          }.execute(tag);

        return tag;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Action findActionByKey (final String key)
      {
        logger.info("==================== performActionByKey(%s)", key);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        final FileObject fo = Repository.getDefault().getDefaultFileSystem().findResource(key);

        if (fo != null && fo.isValid())
          {
            try
              {
                DataObject dob = DataObject.find(fo);
                InstanceCookie ic= dob.getCookie(InstanceCookie.class);

                if (ic != null)
                  {
                    Object instance = ic.instanceCreate();

                    if (instance instanceof Action)
                      {
                        return (Action)instance;
                      }
                  }
              }
            catch (Exception e)
              {
                throw new AssertionError("Can't find action: " + key + " - " + e);
              }
          }

        throw new AssertionError("Can't find action: " + key);
      }

     
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Tag performActionByKey (final String key)
      throws Exception
      {
        logger.info("==================== performActionByKey(%s)", key);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        final Action action = findActionByKey(key);
        final Tag tag = new Tag();
        new AWTExecutor<Void>()
          {
            @Override
            protected Void process()
              {
                action.actionPerformed(null); // FIXME null
                return null;
              }
          }.execute(tag);

        return tag;
      }   

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Dialog performDialogActionByKey (final String key)
      throws Exception
      {
        logger.info("==================== performDialogActionByKey(%s)", key);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        RequestProcessor.getDefault().post(new Runnable()
          {
            public void run()
              {
                try
                  {
                    performActionByKey(key);
                  }
                catch (Exception e)
                  {
                    logger.severe(e.toString());
                    logger.throwing(CLASS, "performDialogActionByKey()", e);
                  }
              }
          });

        return waitForDialog(4000);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Dialog waitForDialog (final long timeout)
      throws Exception
      {
        System.err.println(String.format("===================== waitForDialog(%d)", timeout));
        // Don't wait in the EDT - it's blocked if a modal dialog is running
        waitFor(timeout, 20, new Function()
          {
            public boolean evaluate()
              {
                return DialogRunner.getCurrentModalDialog() != null;
              }
          });

        return DialogRunner.getCurrentModalDialog();
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static void hitKey (final int ... keys)
      {
        logger.info("==================== hitKey(%s)", Arrays.toString(keys));
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        for (int key : keys)
          {
            robot.keyPress(key); 
          }
        
        for (int key : keys)
          {
            robot.keyRelease(key); 
          }

        delay(50);
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Tag clickOn (final JButton button)
      throws InterruptedException, InvocationTargetException
      {
        logger.info("==================== clickOn(%s)", button.getName());
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        return RendezVous.invokeLaterTagged(new Runnable()
          {
            public void run()
              {
                button.doClick();
              }
          });
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Dialog clickAndBringDialog (final JButton button)
      throws Exception
      {
        logger.info("==================== clickAndBringDialog(%s)", button.getName());
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        RequestProcessor.getDefault().post(new Runnable()
          {
            public void run()
              {
                try
                  {
                    clickOn(button);
                  }
                catch (Exception e)
                  {
                    logger.severe(CLASS, e.toString());
                    logger.throwing(CLASS, "clickOnAndBringDialog()", e);
                  }
              }
          });

        return waitForDialog(4000);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Tag setValue (final JSlider slider, final int value)
      throws InterruptedException, InvocationTargetException
      {
        logger.info("==================== setValue(%s)", slider.getName());
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        return RendezVous.invokeLaterTagged(new Runnable()
          {
            public void run()
              {
                slider.setValue(value);
              }
          });
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Tag select (final JToggleButton button, final boolean status)
      throws InterruptedException, InvocationTargetException
      {
        logger.info("==================== select(%s, %s)", button.getName(), status);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        return RendezVous.invokeLaterTagged(new Runnable()
          {
            public void run()
              {
                button.setSelected(status);
              }
          });
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public static Tag select (@Nonnull final Component view,
                              @Nonnull final Node ... nodes)
      throws Exception
      {
        logger.info("==================== select(%s, %s)", view.getName(), Arrays.toString(nodes));
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        final Tag tag = new Tag();
        new AWTExecutor<Void>()
          {
            @Override
            protected Void process()
              throws PropertyVetoException
              {
                ExplorerManager.find(view).setSelectedNodes(nodes);
                return null;
              }
          }.execute(tag);

        return tag;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static void focus (@Nonnull final JComponent component)
      throws Exception
      {
        logger.info("==================== focus(%s)", component.getName());
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        EventQueue.invokeAndWait(new Runnable()
          {
            public void run()
              {
                component.requestFocus();
              }
          });

        delay(1000); // FIXME: why the wait doesn't work
//        waitInAWT(4000, 10, new Function()
//          {
//            public boolean evaluate()
//              {
//                return component.isFocusOwner();
//              }
//          });
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public static Tag select (@Nonnull final JComboBox combobox, @Nonnull final Object item)
      throws Exception
      {
        logger.info("==================== select(%s, %s)", combobox.getName(), item);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        final Tag tag = new Tag();

        new AWTExecutor<Void>()
          {
            @Override
            protected Void process() throws Exception
              {
                for (int i = 0; i < combobox.getItemCount(); i++)
                  {
                    if (item == combobox.getItemAt(i))
                      {
                        combobox.setSelectedIndex(i);
                        return null;
                      }
                  }

                fail("Can't select " + item);
                return null; // never reached
              }
          }.execute(tag);

        return tag;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public static Tag selectByDisplayName (@Nonnull final JComboBox combobox, @Nonnull final String itemDisplayName)
      throws Exception
      {
        logger.info("==================== select(%s, %s)", combobox.getName(), itemDisplayName);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        final Tag tag = new Tag();

        new AWTExecutor<Void>()
          {
            @Override
            protected Void process() throws Exception
              {
                final ListCellRenderer listCellRenderer = combobox.getRenderer();

                for (int i = 0; i < combobox.getItemCount(); i++)
                  {
                    final Object item = combobox.getItemAt(i);
                    // FIXME: doesn't work because of that null
                    final JLabel label = (JLabel)listCellRenderer.getListCellRendererComponent(null, item, i, false, false);

                    if (itemDisplayName.equals(label.getText()))
                      {
                        combobox.setSelectedIndex(i);
                        return null;
                      }
                  }

                fail("Can't select " + itemDisplayName);
                return null; // never reached
              }
          }.execute(tag);

        return tag;
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Tag selectByClassName (@Nonnull final JComboBox combobox, @Nonnull final String itemClassName)
      throws Exception
      {
        logger.info("==================== selectByClassName(%s, %s)", combobox.getName(), itemClassName);        
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());
        
        final Tag tag = new Tag();

        new AWTExecutor<Void>()
          {
            @Override
            protected Void process() throws Exception
              {
                for (int i = 0; i < combobox.getItemCount(); i++)
                  {
                    final Object item = combobox.getItemAt(i);
                    System.err.println("ITEM: " + item);

                    if (itemClassName.equals(item.getClass().getName()))
                      {
                        combobox.setSelectedIndex(i);
                        return null;
                      }
                  }

                fail("Can't select " + itemClassName);
                return null;
              }
          }.execute(tag);

        return tag;
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Tag type (@Nonnull final JTextComponent component, @Nonnull final String text)
      throws Exception
      {
        logger.info("==================== type(%s)", text);
        assertFalse("Called by the EDT", SwingUtilities.isEventDispatchThread());

        final Tag tag = new Tag();

        new AWTExecutor<Void>()
          {
            @Override
            protected Void process()
              {
                component.setText(text);
                return null;
              }
          }.execute(tag);

        return tag;
      }
            
    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull 
    public static File getTestSetPath()
      {
        final File workspaceFolder = Locator.find(WorkspaceManager.class).getWorkspaceFolder();
        return new File(workspaceFolder, "TestSets");
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public static Node[] subArray (@Nonnull final Node[] nodes, final int from, final int length)
      {
        final Node[] result = new Node[length];
        System.arraycopy(nodes, from, result, 0, length);
        return result;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public static Node[] reversed (@Nonnull final Node[] nodes)
      {
        final List<Node> list = new ArrayList<Node>(Arrays.asList(nodes));
        Collections.reverse(list);
        return list.toArray(new Node[0]);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public static Node[] difference (@Nonnull final Node[] a, @Nonnull final Node[] b)
      {
        final List<Node> la = new ArrayList<Node>(Arrays.asList(a));
        la.removeAll(Arrays.asList(b));
        return la.toArray(new Node[0]);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public static <T> List<T> sorted (@Nonnull final List<T> list,
                                      @Nonnull final Comparator<? super T> comparator)
      {
        final List<T> result = new ArrayList<T>(list);
        Collections.sort(result, comparator);
        return result;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public static Node[] findNodes (@Nonnull final JComponent view)
      throws Exception
      {
        return new AWTExecutor<Node[]>()
          {
            @Override
            protected Node[] process()
              {
                // FIXME: Would be nicer to directly get the nodes from the View... is it possible?
                final Node rootNode = ExplorerManager.find(view).getRootContext();
                assertNotNull(rootNode);
                final Node[] nodes = rootNode.getChildren().getNodes();
                return nodes;
              }
          }.execute(null);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Nonnull
    public static Node[] findSelectedNodes (@Nonnull final Component view)
      throws Exception
      {
        return new AWTExecutor<Node[]>()
          {
            @Override
            protected Node[] process()
              {
                // FIXME: Would be nicer to directly get the nodes from the View... is it possible?
                return ExplorerManager.find(view).getSelectedNodes();
              }
          }.execute(null);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Date createDate (final int year, final int month, final int day, final int hour, final int minute, final int second)
      {
        return new Date(year - 1900, month - 1, day, hour, minute, second);  
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    public static Date createDate (final int year, final int month, final int day, final int hour, final int minute, final int second, final int msec)
      {
        return new Date(createDate(year, month, day, hour, minute, second).getTime() + msec);
      }
  }
